/*
 * usb_types.h -- USB Atmel programmer USB data types shared with host 
 *                side implementation. 
 * 
 * Copyright (c) 2005--2008 by Wolfgang WIESER ] wwieser (a) gmx <*> de [
 * 
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#ifndef _AVR_USB_TYPE_H_
#define _AVR_USB_TYPE_H_ 1

// We have pipeline size of this many large packets. 
#define USB_LARGE_PIPELINE_SIZE_ATMELPRG 3
#define USB_LARGE_PIPELINE_SIZE_UNIPROG  6

#ifndef USB_MAIN_DEFINES_ONLY

// NOTE! This is C-code, not C++. 
//       C++ will make USBPacket_ident_req with size 1, C with correct size 0. 
#ifdef __cplusplus
# error "You should not compile this with a C++ compiler."
#endif

#ifndef USB_ON_HOST_SIDE
# error "You need to define USB_ON_HOST_SIDE to 0 or 1."
#endif


// USB commands: 
// Commands 0..15 are reserved by protocol. 
// Commands 16..255 are per-device. 
// IDENTIFY: Request identification from device; 
//           Identification response. 
#define USB_CMD_IDENTIFY     0x01
// ERROR: Response from device for unknown commands. 
#define USB_CMD_ERROR        0x02
// Status request/response (USB-AtmelPrg). 
#define USB_CMD_UAP_STATUS   0x10
// Read bytes. 
#define USB_CMD_UAP_READ     0x11
// Read low/high. 
#define USB_CMD_UAP_READ2    0x12
// Write bytes. 
#define USB_CMD_UAP_WRITE    0x13
// Write low/high. 
#define USB_CMD_UAP_WRITE2   0x14
// Other Atmel commands. 
#define USB_CMD_UAP_ATCMD    0x15
// Waveform output command. 
#define USB_CMD_UAP_WAVEFORM 0x16
// IO status request/response (USB-UniProg). 
#define USB_CMD_UAP_IOSTATUS 0x17


// Size in bytes of the USB packet header (command+length): 
#define USBPPACKET_HEADER_SIZE 2

struct USBPacket_ident_req {}__attribute__((__packed__));

// Note: For nibble-based transfer, the complete packet size may not exceed 
//       127 bytes so that the nibble count still fits into an 8bit integer. 
//       This is not an issue with the packet format but with the internal 
//       counters in the receiving AVR. So, when not using nibbles, we can 
//       transfer up to 255 bytes per packet. 
#define USB_IDENTIFY_DEVICE_NAME_LEN     32
#define USB_IDENTIFY_DEVICE_VENDOR_LEN   32

struct USBPacket_ident_resp
{
	uint8 protocol_version;
	uint8 adapter_type;  // 0 -> USB-AtmelPrg, 1 -> USB-UniProg. 
	uint16 firmware_version;
	char device_name[USB_IDENTIFY_DEVICE_NAME_LEN];
	char device_vendor[USB_IDENTIFY_DEVICE_VENDOR_LEN];
}__attribute__((__packed__));


// NOTE: Packet not understood by USB-UniProg (-> iostatus_req). 
struct USBPacket_uap_status_req
{
	// bits 0..3: 0 -> off; 1 -> on; 0xf -> no change
	// bits 4..7: 0,1 -> disable/enable busy LED; 0xf -> no change
	uint8 set_led;      
	// bit 0,1: 0 -> LOW, 1 -> HIGH, 2 -> High-Z
	//       7: 0 -> ignore; 1 -> change reset signal before SPI change
	//       6: 0 -> ignore; 1 -> change reset signal after SPI change
	uint8 atmel_reset;
	// bit 0,1: 0 -> disable MOSI,SCK (HighZ); 1 -> enable (LOW out), 
	//          2 -> set pullups
	//       2: 0 -> no pullup on MISO; 1 -> pullup on MISO
	//       7: 0 -> ignore; 1 -> change SPI interface status
	uint8 spi_status;
	// Set status of pins 6,7,8. 
	// bit 0: direction 0 -> IN, 1 -> OUT
	//     1: OUT: LOW/HIGH; IN: HighZ/PullUp. 
	// bit 7: 0 -> ignore; 1 -> change
	uint8 sd6,sd7,sd8;
}__attribute__((__packed__));


#define UAP_B_STATUS_LED_ON            0x01U
#define UAP_B_STATUS_MISO_HIGH         0x02U
#define UAP_B_STATUS_BUSY_LED_ENABLED  0x04U

// NOTE: Packet not sent by USB-UniProg (-> iostatus_resp). 
struct USBPacket_uap_status_resp
{
	uint8 portc_val;  // PORTC
	uint8 portc_dir;  // DDRC
	uint8 portc_in;   // inp(PINC)
	uint8 b_status;   // See UAP_B_STATUS_*
}__attribute__((__packed__));


#define UAP_RELAIS_DIO_CLOSE  (1U<<0)
#define UAP_RELAIS_DIO_OPEN   (1U<<1)
#define UAP_RELAIS_5V_CLOSE   (1U<<2)
#define UAP_RELAIS_5V_OPEN    (1U<<3)
#define UAP_RELAIS_LDO_CLOSE  (1U<<4)
#define UAP_RELAIS_LDO_OPEN   (1U<<5)
#define UAP_RELAIS_GND_CLOSE  (1U<<6)
#define UAP_RELAIS_GND_OPEN   (1U<<7)

// NOTE: Packet not understood by USB-AtmelPrg (-> status_req). 
// (11 bytes + header)
struct USBPacket_uap_iostatus_req
{
	// bits 0..3: 0 -> off; 1 -> on; 0xf -> no change
	// bits 4..7: 0,1 -> disable/enable busy LED; 0xf -> no change
	uint8 set_led;      
	// bit 0,1: 0 -> LOW, 1 -> HIGH, 2 -> High-Z
	//       7: 0 -> ignore; 1 -> change reset signal before SPI change
	//       6: 0 -> ignore; 1 -> change reset signal after SPI change
	uint8 atmel_reset;
	// Control the IO pins: 
	// Only change pins which have their bits set in ctrl_mask. 
	uint8 ctrl_mask;
	// Set data direction (0 -> in; 1 -> out) of bits in ctrl_mask. 
	uint8 ddr_bits;
	// Set level (1 -> HIGH; 0 -> LOW) for those bits with ddr=1 and ctrl=1. 
	uint8 level_bits;
	// See UAP_RELAIS_*
	// bit 0 present: close IO relais
	// bit 1 present: open IO relais
	// bit 2 present: close 5V relais
	// bit 3 present: open 5V relais
	// bit 4 present: close LDO relais
	// bit 5 present: open LDO relais
	// bit 6 present: close GND relais
	// bit 7 present: open GND relais
	// Use 0x00 for "no change". 
	uint8 set_relais;
	// Choose IO voltage and SUP voltage. 150..500 in 0.01V. 
	// Use 0000 to shut down the regulator and 0xffff for "no change". 
	uint16 io_voltage;
	uint16 sup_voltage;
}__attribute__((__packed__));

#define UAP_IOSTAT_RELAIS_DIO_CLOSED  (1U<<0)
#define UAP_IOSTAT_RELAIS_5V_CLOSED   (1U<<1)
#define UAP_IOSTAT_RELAIS_LDO_CLOSED  (1U<<2)
#define UAP_IOSTAT_RELAIS_GND_CLOSED  (1U<<3)
#define UAP_IOSTAT_LDO_VCCIO_ENABLED  (1U<<4)
#define UAP_IOSTAT_LDO_SUP_ENABLED    (1U<<5)
#define UAP_IOSTAT_LED_ON             (1U<<6)
#define UAP_IOSTAT_BUSY_LED_ENABLED   (1U<<7)

// NOTE: Packet not sent by USB-AtmelPrg (-> status_resp). 
// (12 bytes + header)
struct USBPacket_uap_iostatus_resp
{
	// Bitbang port state. Note that when the JTAG system in the CPLD is 
	// active, some output ports can be overridden. 
	uint8 bb_ddr,bb_level,bb_in;
	// Relais status + LDO shutdown status (6 bits) + LED status (2 bits). 
	// (busy LED enabled? LED on?)
	// See UAP_IOSTAT_*
	uint8 status;
	// Measured voltages in mV (4x16 bits). 
	uint16 adc_val[4];
}__attribute__((__packed__));


// Enter programming mode. 
//  Input: v8: 0/1: try_clock_shift value
//  Output: v32: response from _ATSerialIO32() 
//          status: 0 -> OK; 1 -> failed 
#define UAP_ATCMD_ENTER_PROG_MODE 0x01U
// Read signature. 
//  Input: no additional params
//  Output: v32: vendor<<16 | flash_size<<8 | device
//          status: always 0
#define UAP_ATCMD_READ_SIGNATURE  0x02U
// Execute arbitrary single command. 
//  Input: v32: command (like 0x58000000U for "read lock") 
//  Output: status: read back status from _ATSerialIO() 
#define UAP_ATCMD_EXEC_SINGLE_CMD 0x03U
// Set IO speed for communication with device to be programmed. 
//  Input: v8: speed value (0: 0.6usec; 1: 1.125usc, 2:2.125usec, >=3: delay)
//      (Exact delay specs for 0,1,2 differ somewhat for UniProg and AtmelPrg.)
//  Output: status: v8 value actually used. 
#define UAP_ATCMD_SET_IO_SPEED    0x04U

struct USBPacket_uap_atcmd_req
{
	uint8 cmd;    // UAP_ATCMD_*
	uint8 v8;     // 8bit value depending on cmd. 
	uint32 v32;   // 32bit value depending on cmd. 
}__attribute__((__packed__));

struct USBPacket_uap_atcmd_resp
{
	uint8 cmd;     // Command echoed back from request. 
	uint8 status;  // Status; depending on command. 
	uint32 v32;    // 32bit arbitrary value depending on command. 
}__attribute__((__packed__));


// Read: read bytes
struct USBPacket_uap_read_req
{
	uint16 adr;    // start address in 1byte words
	uint32 cmd;    // command for reading
	uint8  len;    // number of 1byte words to read; max. 64 (64 bytes)
}__attribute__((__packed__));

// Read2: read low and high alternatively. 
struct USBPacket_uap_read2_req
{
	uint16 adr;    // start address in 2byte words
	uint32 cmd_L;  // command for reading LOW
	uint32 cmd_H;  // command for reading HIGH
	uint8  len;    // number of 2byte words to read; max. 32 (64 bytes)
}__attribute__((__packed__));

struct USBPacket_uap_data_resp
{
	union {
		uint16 data16[32];  // always length 64 bytes 
		uint8  data8[64];   // always length 64 bytes 
	};
}__attribute__((__packed__));


// Write: write bytes
struct USBPacket_uap_write_req
{
	uint16 adr;    // start address in 1byte words
	uint32 cmd_W;  // command for writing
	uint32 cmd_R;  // command for reading (!)
	uint8  len;    // number of 1byte words to write; max. 64 (64 bytes)
	uint8  delay;  // fallback delay in msec (normally use 6)
	uint8  data8[64];   // data to write (always length 64 bytes) 
}__attribute__((__packed__));

// Write: write low and high alternatively. 
// For paged writing, use paging!=0. 
//  First, all passed data will be loaded into the page buffer. 
//  The cmd_WP will be used to write the page if paging=2 (but not for 
//  paging=1). 
struct USBPacket_uap_write2_req
{
	uint16 adr;     // start address in 1byte words
	uint32 cmd_WL;  // command for writing LOW
	uint32 cmd_WH;  // command for writing HIGH
	uint8  len;     // number of 1byte words to write; max. 64 (64 bytes)
	uint8  paging;  // paged writing or one byte at a time?
	uint8  delay;   // fallback delay in msec (use 6 for paging=0, else 15)
	union {
		// For paging=0: 
		struct {
			uint8 _pad;
			uint32 cmd_RL;  // command for reading LOW
			uint32 cmd_RH;  // command for reading HIGH
		}__attribute__((__packed__));
		struct {
			uint8  poll_byte;  // data byte for data polling; 
		    	               // 0xff for msec delay in delay. 
			uint32 cmd_WP;  // command to write page
			uint32 cmd_P;   // command for data polling
		}__attribute__((__packed__));
	}__attribute__((__packed__));
	uint16 data16[32];   // data to write (always length 64 bytes) 
}__attribute__((__packed__));

// Sent back for write and write2 reqs. 
struct USBPacket_uap_write_resp
{
	uint8 status;
}__attribute__((__packed__));


// Waveform output request. 
// wave_mode=0: (USB-AtmelPrg)
//   With a constant clock rate, the passed data is written to output pins: 
//   bits 0,1,2,3 are emitted on SCK,MOSI,RESET,PC3, respectively, then 
//   bits 4,5,6,7 are emitted on SCK,MOSI,RESET,PC3, respectively. 
//   Hence, with 64 bytes (max allowed size), 128 state changes can be achieved. 
//   (I.e. wave_size<=128) 
//   After each state change, MISO is read in and stored in the response 
//   packet making up 16 bytes max (i.e. resp_bits<=128). 
// wave_mode=1: (USB-AtmelPrg)
//   With constant clock rate, the passed wave_data is emitted on the 
//   pins MOSI and RESET in the following way: 
//   bits 0,1 are put on MOSI,RESET while taking SCK low, then SCK is taken high, 
//   then bits 2,3 are put on MOSI,RESET while taking SCK low, then SCK high, 
//   then bits 4,5....
//   In the end, SCK is high. 
//   With 64 bytes wave_data, max. 256 state changes can be passed 
//   (wave_size<=256). 
//   The MISO response is read after the falling edge of SCK (shortly before or 
//   after taking SCK high). 
//   At most 256 bits of MISO data are read (resp_bits<=256). 
// wave_mode=2: (USB-UniProg)
//   With constant clock rate, the passed wave_data is emitted on the 
//   pins MOSI/TDI and RESET/TMS in the following way: 
//   data[0] is loaded into the TDI shift register, data[1] into the TMS shift 
//   register, then up to 2*8 bits are shifted out. Then data[2] and data[3] 
//   are loaded and so on. Bits are shifted out MSB first. 
// All other modes: no-op, 0 bits of response. 
struct USBPacket_uap_waveform_req
{
	uint8 wave_mode;      // waveform output operation mode; see above
	uint8 delay;          // wave emission speed/delay value 
	uint16 wave_size;     // number of state changes
	uint8 wave_data[64];  // see above
}__attribute__((__packed__));

struct USBPacket_uap_waveform_resp
{
	uint16 resp_bits;      // number of bits in response
	uint8 miso_data[64];   // reserve 64 bytes to have some room for expansion
	                       // (same size as SPI read responses)
}__attribute__((__packed__));


// USB protocol (send) packet. 
#if USB_ON_HOST_SIDE
struct USBRecvPacket
#else
struct USBSendPacket
#endif
{
	uint8 command;   // Command; see USB_CMD_*
	uint8 length;    // Data packet length including command. 
#if USB_ON_HOST_SIDE
	union {
		struct USBPacket_ident_resp ident_resp;
		struct USBPacket_uap_status_resp uap_status_resp;
		struct USBPacket_uap_iostatus_resp uap_iostatus_resp;
		struct USBPacket_uap_data_resp uap_data_resp;
		struct USBPacket_uap_atcmd_resp uap_atcmd_resp;
		struct USBPacket_uap_write_resp uap_write_resp;
		struct USBPacket_uap_waveform_resp uap_waveform_resp;
	}__attribute__((__packed__));
#else
	uint8 uap_buf[];  // -> ext. buf
#endif
}__attribute__((__packed__));


// USB protocol (recv) packet. 
#if USB_ON_HOST_SIDE
struct USBSendPacket
#else
struct USBRecvPacket
#endif
{
	uint8 command;   // Command; see USB_CMD_*
	uint8 length;    // Data packet length including command. 
	union {
		struct USBPacket_ident_req ident_req;
		//----------------------------------------------------------------------
		struct USBPacket_uap_status_req uap_status_req;
		struct USBPacket_uap_iostatus_req uap_iostatus_req;
		struct USBPacket_uap_read_req uap_read_req;
		struct USBPacket_uap_read2_req uap_read2_req;
		struct USBPacket_uap_write_req uap_write_req;
		struct USBPacket_uap_write2_req uap_write2_req;
		struct USBPacket_uap_atcmd_req uap_atcmd_req;
		struct USBPacket_uap_waveform_req uap_waveform_req;
	}__attribute__((__packed__));
}__attribute__((__packed__));

#endif  /* USB_MAIN_DEFINES_ONLY */
#endif  /* _AVR_USB_TYPE_H_ */
