/*
 * utils.c -- USB Atmel programmer utility functions. 
 * 
 * Copyright (c) 2005--2006 by Wolfgang Wieser ] wwieser (a) gmx <*> de [
 * 
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <time.h>

#include <ftdi.h>
#include "prototypes.h"


int ReadEnvVar_FTDI_USB(int *vendor,int *device)
{
	// Check for vendor and device in the FTDI_USB env var. 
	const char *envvar=getenv("FTDI_USB");
	if(envvar)
	{
		int v=*vendor,d=*device;
		if(sscanf(envvar,"%x%*c%x",&v,&d)<2)
		{
			fprintf(stderr,"Failed to parse FTDI_USB env var (%s). "
				"Falling back to defaults.\n",
				envvar);
			return(-1);
		}
		else
		{
			*vendor=v;
			*device=d;
			return(0);
		}
	}
	
	return(1);
}


int InstallSignalHandler(int sig,void (*shandler)(int),int sa_flags)
{
	struct sigaction act;
	memset(&act,0,sizeof(act));
	sigemptyset(&act.sa_mask);
	act.sa_handler=shandler;
	act.sa_flags=sa_flags;
	
	return(sigaction(sig,&act,NULL));
}


long msec_elapsed(const struct timeval *old,const struct timeval *current)
{
	struct timeval _curr;
	if(!current)
	{
		gettimeofday(&_curr,NULL);
		current=&_curr;
	}
	return(
		(current->tv_sec  - old->tv_sec )*1000L + 
		(current->tv_usec - old->tv_usec)/1000L );
}


// Reverse of ftdi_eeprom_build() from libftdi: 
void ftdi_eeprom_unbuild(unsigned char *buf,struct ftdi_eeprom *eeprom)
{
	// Addr 00: Stay 00 00
	// Addr 02: Vendor ID
	eeprom->vendor_id = (int)buf[0x02] | (int)buf[0x03]<<8;
	
	// Addr 04: Product ID
	eeprom->product_id = (int)buf[0x04] | (int)buf[0x05]<<8;
	
	// Addr 06: Device release number (0400h for BM features)
	switch((int)buf[0x06] | (int)buf[0x07]<<8)
	{
		case 0x0400:  eeprom->BM_type_chip=1;  break;
		case 0x0200:  eeprom->BM_type_chip=0;  break;
		default:  eeprom->BM_type_chip=-1;  break;
	}
	
	// Addr 08: Config descriptor
	// Bit 1: remote wakeup if 1
	// Bit 0: self powered if 1
	//
	eeprom->self_powered = (buf[0x08]&0x01) ? 1 : 0;
	eeprom->remote_wakeup = (buf[0x08]&0x02) ? 1 : 0;
	
	// Addr 09: Max power consumption: max power = value * 2 mA
	eeprom->max_power = buf[0x09];
	
	// Addr 0A: Chip configuration
	// Bit 7: 0 - reserved
	// Bit 6: 0 - reserved
	// Bit 5: 0 - reserved
	// Bit 4: 1 - Change USB version
	// Bit 3: 1 - Use the serial number string
	// Bit 2: 1 - Enable suspend pull downs for lower power
	// Bit 1: 1 - Out EndPoint is Isochronous
	// Bit 0: 1 - In EndPoint is Isochronous
	//
	eeprom->in_is_isochronous = (buf[0x0A]&0x01) ? 1 : 0;
	eeprom->out_is_isochronous = (buf[0x0A]&0x02) ? 1 : 0;
	eeprom->suspend_pull_downs = (buf[0x0A]&0x04) ? 1 : 0;
	eeprom->use_serial = (buf[0x0A]&0x08) ? 1 : 0;
	eeprom->change_usb_version = (buf[0x0A]&0x10) ? 1 : 0;
	
	// Addr 0B: reserved
	
	// Addr 0C: USB version low byte when 0x0A bit 4 is set
	// Addr 0D: USB version high byte when 0x0A bit 4 is set
	//if (eeprom->change_usb_version == 1)
		eeprom->usb_version = (int)buf[0x0C] | (int)buf[0x0D]<<8;
	
	eeprom->manufacturer=NULL;
	eeprom->product=NULL;
	eeprom->serial=NULL;
	
	do {
		// Addr 0E: Offset of the manufacturer string + 0x80
		int manufacturer_offset = (int)buf[0x0E]-0x80;
		
		// Addr 0F: Length of manufacturer string
		int manufacturer_size=(int)buf[0x0F]/2-1;
		
		// Addr 10: Offset of the product string + 0x80
		int product_offset = (int)buf[0x10]-0x80;
		
		// Addr 11: Length of product string
		int product_size=(int)buf[0x11]/2-1;
		
		// Addr 12: Offset of the serial string + 0x80
		int serial_offset = (int)buf[0x12]-0x80;
		
		// Addr 13: Length of serial string
		int serial_size = (int)buf[0x13]/2-1;
		
		int dummy,i,j;
		
		// Read manufacturer
		if(manufacturer_offset+2>=128)  break;
		dummy=(int)buf[manufacturer_offset]/2-1;
		if(dummy!=manufacturer_size)  break;
		if(buf[manufacturer_offset+1]!=0x03)  break;  // type: string
		
		i = manufacturer_offset+2;
		eeprom->manufacturer=(char*)malloc(manufacturer_size+2);
		if(i+manufacturer_size*2>=128)  break;
		for (j = 0; j < manufacturer_size; j++,i+=2)
			eeprom->manufacturer[j]=buf[i];
		eeprom->manufacturer[j]='\0';
		
		// Read product name
		if(product_offset+2>=128)  break;
		dummy=(int)buf[product_offset]/2-1;
		if(dummy!=product_size)  break;
		if(buf[product_offset+1]!=0x03)  break;  // type: string
		
		i = product_offset+2;
		eeprom->product=(char*)malloc(product_size+2);
		if(i+product_size*2>=128)  break;
		for (j = 0; j < product_size; j++,i+=2)
			eeprom->product[j]=buf[i];
		eeprom->product[j]='\0';
		
		// Read serial
		if(serial_offset+2>=128)  break;
		dummy=(int)buf[serial_offset]/2-1;
		if(dummy!=serial_size)  break;
		if(buf[serial_offset+1]!=0x03)  break;  // type: string
		
		i = serial_offset+2;
		eeprom->serial=(char*)malloc(serial_size+2);
		if(i+serial_size*2>=128)  break;
		for (j = 0; j < serial_size; j++,i+=2)
			eeprom->serial[j]=buf[i];
		eeprom->serial[j]='\0';
	} while(0);
}


void ftdi_eeprom_free(struct ftdi_eeprom *eeprom)
{
	if(!eeprom)  return;
	
	if(eeprom->manufacturer)
	{  free(eeprom->manufacturer);  eeprom->manufacturer=NULL;  }
	if(eeprom->product)
	{  free(eeprom->product);  eeprom->product=NULL;  }
	if(eeprom->serial)
	{  free(eeprom->serial);  eeprom->serial=NULL;  }
}
