/*
 * waveformbuf.cc -- USB Atmel programmer: Waveform buffer for USB IO. 
 * 
 * Copyright (c) 2006--2007 by Wolfgang Wieser ] wwieser (a) gmx <*> de [
 * 
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "config.h"
#include "waveformbuf.h"

#include <stdio.h>

// We need this for USB_LARGE_PIPELINE_SIZE: 
#define USB_MAIN_DEFINES_ONLY 1
#include "usb_types.h"


void *WaveFormBuffer::WaveBlock::operator new(size_t size)
{
	void *ptr=malloc(size);
	assert(ptr);
	return(ptr);
}

void WaveFormBuffer::WaveBlock::operator delete(void *ptr)
{
	if(ptr)  free(ptr);
}


WaveFormBuffer::WaveBlock::WaveBlock() : next(NULL),wave_size(0)
{
	memset(expect_mask,0,32);
	memset(expect_buf,0,32);
	memset(dest,0,256*sizeof(void*));
	memset(dest_bidx,0,256);
}


//------------------------------------------------------------------------------

int WaveFormBuffer::_DoStoreCompareReadback(WaveBlock *wb_recv,char jverbose)
{
	int nerrors=0;
	
	// Store readback. 
	for(int i=0; i<wb_recv->wave_size; i++)
	{
		void *dest=wb_recv->dest[i];
		if(!dest)  continue;
		uint8 bidx=wb_recv->dest_bidx[i];
		uint8 btype=bidx>>6;
		bidx&=63U;
		// Bit is the readback; either 0 or non-0. 
		uint8 bit = wb_recv->resp_buf[i/8] & (0x01U <<(7-(i&0x7)));
		switch(btype)
		{
			case 0:
				assert(bidx<8U);
				if(bit)  *((uint8*)dest) |= ((uint8)0x01U)<<bidx;
				else     *((uint8*)dest) &= ~(((uint8)0x01U)<<bidx);
				break;
			case 1:
				assert(bidx<16U);
				if(bit)  *((uint16*)dest) |= ((uint16)0x01U)<<bidx;
				else     *((uint16*)dest) &= ~(((uint16)0x01U)<<bidx);
				break;
			case 2:
				assert(bidx<32U);
				if(bit)  *((uint32*)dest) |= ((uint32)0x01U)<<bidx;
				else     *((uint32*)dest) &= ~(((uint32)0x01U)<<bidx);
				break;
			case 3:
				assert(bidx<64U);
				if(bit)  *((uint64*)dest) |= ((uint64)0x01U)<<bidx;
				else     *((uint64*)dest) &= ~(((uint64)0x01U)<<bidx);
				break;
			default:  assert(0);
		}
	}

	// Compare readback. 
	int iend=(wb_recv->wave_size+7)/8;
	for(int i=0; i<iend; i++)
	{
		if((wb_recv->resp_buf[i] & wb_recv->expect_mask[i]) != 
			wb_recv->expect_buf[i])
		{
			if(jverbose=='s' && !nerrors)
			{  printf("\n");  fflush(stdout);  }
			fprintf(stderr,"OOPS: read 0x%02x, expect 0x%02x, "
				"mask 0x%02x  (byte %d/%d)\a\n",
				wb_recv->resp_buf[i],wb_recv->expect_buf[i],
				wb_recv->expect_mask[i],
				i,iend);
			++nerrors;
			//getchar();
		}
	}
	
	return(nerrors);
}


int WaveFormBuffer::flush(uint8 wave_delay,char jverbose)
{
	// Quit early if there is nothing to do. 
	if(!send_list)  return(0);
	
	// We count errors; negative values are USB (transfer) errors. 
	int nerrors=0;
	
	//fprintf(stderr,"Flushing (%d blocks) ",pending_blocks);
	
	/*printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
		"Programming SVF file: %d bits in %d blocks... "
		"                         \b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b",
		totsize_sent,nblocks_sent);
	fflush(stdout);*/
	
#if 1  /* PIPELINING */
	const int pipeline_size=USB_GetPipelineSize();
	int in_pipeline=0;
	int dump_cnt=0;
	WaveBlock *wb_pipe[pipeline_size];
	for(int i=0; i<pipeline_size; i++)
	{  wb_pipe[i]=NULL;  }
	
	// Process all the waves in the list. 
	WaveBlock *wb_send=send_list;
	// Sigint is caught when waiting for response and will result in error 
	// return. 
	while((wb_send || in_pipeline) /*&& !caught_sigint*/)
	{
		if(wb_send)
		{
			assert(wb_send->wave_size>0);
			assert(wb_send->wave_size<=WaveBlockMaxSize);
			
			//fprintf(stderr,"Sending waveform: size=%d [%d]\n",
			//	wb_send->wave_size,in_pipeline);
			USB_Status rv=USB_RawWriteWaveformReq(/*wave_mode=*/1,
				wb_send->wave_size,wb_send->wave_buf,wave_delay);
			if(rv)
			{
				if(jverbose=='s') printf("\n");
				fprintf(stderr,"Error sending waveform req (size=%d): %s\n",
					wb_send->wave_size,USB_Status_String(rv));
				nerrors=-10;
				break;
			}
			
			++nblocks_sent;
			totsize_sent+=wb_send->wave_size;
			
			assert(in_pipeline<pipeline_size);
			wb_pipe[in_pipeline++]=wb_send;
			
			wb_send=wb_send->next;
		}
		
		if(in_pipeline>=pipeline_size || !wb_send)
		{
			WaveBlock *wb_recv=wb_pipe[0];
			
			//fprintf(stderr,"Expecting waveform: size=%d (%d)\n",
			//	wb_recv->wave_size,in_pipeline);
			USB_Status rv=USB_RawReadWaveformResp(wb_recv->wave_size,
				wb_recv->resp_buf);
			if(rv)
			{
				if(jverbose=='s') printf("\n");
				fprintf(stderr,"Error receiving waveform resp: %s\n",
					USB_Status_String(rv));
				nerrors=-11;
				break;
			}
			
			// Shift pointers: 
			for(int i=1; i<in_pipeline; i++)
			{  wb_pipe[i-1]=wb_pipe[i];  }
			wb_pipe[--in_pipeline]=NULL;
			
			nerrors+=_DoStoreCompareReadback(wb_recv,jverbose);
		}
		
		//fprintf(stderr,"---------<%d %d>------------\n",in_pipeline,nerrors);
		
		if(jverbose=='s' && (dump_cnt++==3 || (!wb_send && !in_pipeline)))
		{
			printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
				"Programming SVF file: %d bits in %d blocks... "
				"                    \b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b",
				totsize_sent,nblocks_sent);
			fflush(stdout);
			dump_cnt=0;
		}
		
		//fprintf(stderr,"%d",in_pipeline);
	}
	
	//fprintf(stderr,"---------<%d %d>------------\n",in_pipeline,nerrors);
	
#else  /* SEQUENTIAL */
	// Process all the waves in the list. 
	for(WaveBlock *wb=send_list; wb; wb=wb->next)
	{
		assert(wb->wave_size>0);
		assert(wb->wave_size<=WaveBlockMaxSize);
		
		//fprintf(stderr,"Sending waveform: size=%d\n",wb->wave_size);
		USB_Status rv=USB_RawWaveformIO(/*wave_mode=*/1,wb->wave_size,
			wb->wave_buf,wb->resp_buf);
		++nblocks_sent;
		totsize_sent+=wb->wave_size;
		
		if(rv)
		{
			if(jverbose=='s') printf("\n");
			fprintf(stderr,"Error transmitting waveform: %s\n",
				USB_Status_String(rv));
			nerrors=-10;
			break;
		}
		
		// Compare readback. 
		int iend=(wb->wave_size+7)/8;
		for(int i=0; i<iend; i++)
		{
			nerrors+=_DoStoreCompareReadback(wb,jverbose);
			
			if(caught_sigint) break;
		}
		
		if(caught_sigint) break;
		
		if(jverbose=='s')
		{  fprintf(stderr,".");  }
	}
	
	if(jverbose=='s')
	{  fprintf(stderr,"\n");  }
#endif
	
	clear();
	
	return(nerrors);
}


void WaveFormBuffer::_RemoveFirst()
{
	// Remove first block from queue. 
	WaveBlock *wb=send_list;
	if(list_head==send_list)
	{  list_head=NULL;  }
	send_list=send_list->next;
	delete wb;
	--pending_blocks;
}


void WaveFormBuffer::clear(int /*force_free*/)
{
	while(send_list)
	{
		WaveBlock *wb=send_list;
		send_list=send_list->next;
		delete wb;
		--pending_blocks;
	}
	list_head=NULL;
	assert(pending_blocks==0);
}


void WaveFormBuffer::_append(uint8 bits,uint8 expect,void *dest,uint8 dest_bidx)
{
	if(!list_head)
	{
		assert(!send_list);
		list_head=send_list=new WaveBlock();
		++pending_blocks;
	}
	else if(list_head->wave_size>=WaveBlockMaxSize)
	{
		assert(!list_head->next);
		list_head->next=new WaveBlock();
		list_head=list_head->next;
		++pending_blocks;
	}
	
	// Handle data state: 
	uint8 *d = &list_head->wave_buf[list_head->wave_size/4];
	int bi = list_head->wave_size & 0x03;
	if(bi)  {  *d|=(bits&0x03U)<<(2*bi);  }
	else    {  *d=bits&0x03U;  }
	
	// Handle expected readback values. 
	// The constructor sets up mask and buf to 0, so we only care about cases 
	// where the mask is !=0. 
	if((expect & 0x01U))
	{
		int i=list_head->wave_size;
		list_head->expect_mask[i/8] |= 0x01U <<(7-(i&0x7));
		if((expect & 0x02U))
		{  list_head->expect_buf[i/8] |= 0x01U <<(7-(i&0x7));  }
	}
	
	if(dest)
	{
		list_head->dest[list_head->wave_size]=dest;
		list_head->dest_bidx[list_head->wave_size]=dest_bidx;
	}
	
	++list_head->wave_size;
}


WaveFormBuffer::WaveFormBuffer()
{
	send_list=NULL;
	list_head=NULL;
	
	pending_blocks=0;
	
	nblocks_sent=0;
	totsize_sent=0;
}
