/*
 * buffer.cc -- Buffers as used in USB Atmel programmer. 
 * 
 * Copyright (c) 2003--2005 by Wolfgang Wieser ] wwieser (a) gmx <*> de [
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "config.h"
#include "buffer.h"
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>


void *DataBuffer::operator new(size_t size)
{
	void *ptr=malloc(size);
	assert(ptr);
	return(ptr);
}

void DataBuffer::operator delete(void *ptr)
{
	if(ptr)  free(ptr);
}


void DataBuffer::Clear()
{
	if(buf)
	{  free(buf);  buf=NULL;  }
	size=0;
	allocsize=0;
}

void DataBuffer::Alloc(size_t size)
{
	Clear();
	buf=(char*)malloc(size);
	assert(buf);
	allocsize=size;
}


#if 0
int DataBuffer::ReadEEPROM(AtmelIO *atio,size_t adr,size_t len)
{
	Alloc(len);
	
	printf("Reading EEPROM into buffer \"%s\" (%u bytes)... ",
		name,len);  fflush(stdout);
	atio->ReadEEPROM(adr,len,(uchar*)buf);
	size=len;
	printf(caught_signal ? "[signal]\n" : "done\n");  fflush(stdout);
	
	return(0);
}

int DataBuffer::WriteEEPROM(AtmelIO *atio,size_t adr)
{
	printf("Writing EEPROM from buffer \"%s\" (%u bytes)... ",
		name,size);  fflush(stdout);
	atio->WriteEEPROM(adr,size,(uchar*)buf);
	printf(caught_signal ? "[signal]\n" : "done\n");  fflush(stdout);
	
	return(0);
}
#endif


int DataBuffer::ReadBinaryFile(const char *file)
{
	Clear();  // to be sure
	
	printf("Reading binary file %s into buffer \"%s\"... ",
		file,name);  fflush(stdout);
	
	int fd=open(file,O_RDONLY);
	if(fd<0)
	{  fprintf(stderr,"Failed to open %s: %s\n",file,strerror(errno));
		return(-2);  }
	
	for(;;)
	{
		if(allocsize<size+4096)
		{
			allocsize+=4096;
			buf=(char*)realloc(buf,allocsize);
			assert(buf);
		}
		ssize_t rd=read(fd,buf+size,4096);
		if(rd<0)
		{  fprintf(stderr,"Reading %s: %s\n",file,strerror(errno));
			close(fd);  return(-3);  }
		if(!rd)  break;
		size+=rd;
	}
	
	close(fd);
	
	printf("%u bytes\n",size);  fflush(stdout);
	
	return(0);
}


int DataBuffer::WriteBinaryFile(const char *file)
{
	printf("Writing buffer \"%s\" into binary file %s (%u bytes)... ",
		name,file,size);  fflush(stdout);
	
	int fd=open(file,O_WRONLY|O_CREAT|O_TRUNC,0666);
	if(fd<0)
	{  fprintf(stderr,"Failed to open %s: %s\n",file,strerror(errno));
		return(-2);  }
	
	size_t written=0;
	while(written<size)
	{
		ssize_t wr=write(fd,buf+written,size-written);
		if(wr<0)
		{  fprintf(stderr,"Writing %s: %s\n",file,strerror(errno));
			close(fd);  return(-3);  }
		written+=wr;
	}
	
	close(fd);  // <-- should check close errors...
	
	printf("OK\n");  fflush(stdout);
	
	return(0);
}


static inline void _HexdumpPutChar(FILE *out,unsigned char x)
{
	if(isprint(x))
	{  fprintf(out,"%c",x);  }
	else
	{  fprintf(out,".");  }
}


void DataBuffer::HexDumpBuffer(FILE *out,int with_ascii)
{
	fprintf(out,"Dumping buffer \"%s\" (%u bytes) [16 words/line LLHH]\n",
		name,size);
	u_int16_t *data=(u_int16_t*)buf;
	size_t dsize=size/2;
	ssize_t skip_start=-1;
	for(size_t i=0; i<dsize; )
	{
		// Do not dump lines which are only ffffff....
		size_t j=0;
		int do_skip=1;
		for(size_t ii=i; j<16 && ii<dsize; j++,ii++)
		{
			if(data[ii]!=0xffffU)
			{  do_skip=0;  break;  }
		}
		if(do_skip)
		{
			if(skip_start<0)  skip_start=i;
			i+=16;
			continue;
		}
		else if(skip_start>=0)
		{
			printf("         [skipping 0x%04x..0x%04x: "
				"%u words 0xffff]\n",
				size_t(skip_start),i-1,i-size_t(skip_start));
			skip_start=-1;
		}
		
		printf("  0x%04x ",i);
		size_t oldi=i;
		for(j=0; j<16 && i<dsize; j++,i++)
		{
			if(j && !(j%8))  printf(" ");
			printf("%02x%02x",data[i]&0xffU,data[i]>>8);
		}
		// This adds a plaintext column (printable chars only): 
		if(with_ascii)
		{
			for(; j<16; j++)
			{  printf((j && !(j%8)) ? "     " : "    ");  }
			
			printf("    ");
			i=oldi;
			for(j=0; j<16 && i<dsize; j++,i++)
			{
				_HexdumpPutChar(out,data[i]&0xffU);
				_HexdumpPutChar(out,data[i]>>8);
			}
		}
		printf("\n");
	}
	if(skip_start>=0 && dsize>0)
	{
		printf("         [skipping 0x%04x..0x%04x: "
			"%u words 0xffff]\n",
			size_t(skip_start),dsize-1,dsize-size_t(skip_start));
		skip_start=-1;
	}
	fflush(out);
}


DataBuffer::DataBuffer(const char *_name)
{
	name=strdup(_name);
	assert(name);
	buf=NULL;
	size=0;
	allocsize=0;
	next=NULL;
}

DataBuffer::~DataBuffer()
{
	Clear();
	if(name)
	{  free(name);  name=NULL;  }
}


/******************************************************************************/

DataBuffer *DataBufferList::GetDataBuffer(const char *name,int do_create)
{
	for(DataBuffer *i=head; i; i=i->next)
	{
		if(!strcmp(i->name,name))
		{  return(i);  }
	}
	if(!do_create)
	{  return(NULL);  }
	DataBuffer *b=new DataBuffer(name);
	b->next=head;
	head=b;
	return(b);
}


int DataBufferList::RemoveDataBuffer(const char *name)
{
	for(DataBuffer *p=NULL,*i=head; i; p=i,i=i->next)
	{
		if(strcmp(i->name,name))  continue;
		// Remove buffer *i. 
		if(p)
		{  p->next=i->next;  }
		else
		{  head=i->next;  }
		delete i;
		return(0);
	}
	
	return(1);
}


void DataBufferList::Clear()
{
	while(head)
	{
		DataBuffer *tmp=head;
		head=head->next;
		delete tmp;
	}
}


DataBufferList::DataBufferList()
{
	head=NULL;
}

DataBufferList::~DataBufferList()
{
	Clear();
}
