/*
 * main.cc -- USB Atmel/Universal programmer main file. 
 * 
 * Copyright (c) 2005--2009 by Wolfgang Wieser ] wwieser (a) gmx <*> de [
 * 
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation. (See COPYING.GPL for details.)
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

// FIXME: Store io,... voltages in programmer
//        sup 0 / off must force a "disconnect" so that e.g. a "getfl" will 
//             connect first
//        rl/rh... must make sure the relais are connected!

#include "config.h"
#include "adapter.h"
#include "../cable/usbio.h"
#include "buffer.h"
#include "atmel.h"
#include "jtag.h"
#include "../cable/eeprom.h"
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/poll.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>


#if HAVE_JTAG_SYSTEM
// Interface to JTAG system: 
extern "C" {
	extern int jtag_system_start( void );
	extern int jtag_system_parse_line(char *line);
	extern void jtag_system_cleanup( void );
}
#endif

#if HAVE_LIBREADLINE
# include <readline/readline.h>
# include <readline/history.h>
#endif

#if !defined(VERSION) || !defined(VERSDATE)
# error Need to define VERSION and VERSDATE. 
#endif

#define FTDI_USB_VENDOR_CODE      0x0403
#define FTDI_DEVICE_ATMELPRG      0x6002  /* 6002 for USB-AtmelPrg */
#define FTDI_DEVICE_UNIPROG       0x6004  /* 6004 for USB-UniProg */
#define FTDI_UNCONFIGURED_DEVICE  0x6001  /* unconfigured FTDI USB SIO */

//----------------------------CONFIG--------------------------------------
// Poll USB input every these many msec when NOT expecting data. 
static int idle_poll_msec=500;
// Poll USB input every these many usec when expecting data. 
// Fastest operation is achieved with delay 0. The actual delay should 
// be below 10ms. 
int busy_poll_usec=500;  // 500

// USB response timeout in msec. 
int usb_resp_timeout=8000;

// Is the reset pin for the device to program externally controlled (i.e. 
// use High-Z state on it) or do we control it?
static int external_reset=0;

// Try single SCK clock shifts during connect?
static uint8 try_clock_shift=0;

// IO speed for communication with device to be programmed. 
static uint8 spi_io_speed=2;  // "normal"
// JTAG speed. Gets sent with waveform packet. 
//  USB-AtmelPrg: Speed is 1MHz/(jtag_speed+1). 
//  USB-UniProg:  Speed is 12MHz/jtag_speed and 24MHz if jtag_speed==0. 
uint8 usbatmelprg_jtag_speed=0;

// Make JTAG RUNTEST delays longer by this factor/10. 
// Hence a delay d becomes d + jtag_delay_prolong_fact*d/10;
// Xilinx CPLDs make trouble when this is set to 0. Be conservative. 
int jtag_delay_prolong_fact=5;

//------------------------------------------------------------------------

// Name this program was called without path. 
const char *prg_name=NULL;
// Readline, etc. prompt name. 
static char prompt_str[32]={0};

// Current settings. Set defaults here. 
static int curr_io_voltage=330;
static int curr_sup_voltage=330;
// If the voltages are valid, i.e. set like that in the programmer. 
static int curr_io_voltage_valid=0;
static int curr_sup_voltage_valid=0;

// Are we connected to a supported board?
// Use if(USB_GetAdapterType()!=AT_None). 

// Verbosity level. 
int verbose=0;

volatile int caught_sigint=0;
static volatile int got_eof=0;

// List of all buffers. 
DataBufferList buf_list;

// Connected Atmel chip which we want to program or NULL before connect/after 
// disconnect. 
static AtmelDevice *atmel=NULL;

#if HAVE_JTAG_SYSTEM
// Main operation mode. 
static enum OpMode
{
	MAINSYS,
	JTAGSYS,
} opmode=MAINSYS;
#endif


// Signal handler for ^C/SIGINT. 
void _caught_sigint(int)
{
	++caught_sigint;
}

#undef MIN
template<typename T>static inline T MIN(T a,T b)
	{  return(a<b ? a : b);  }


static void PrintUsage(int exit_val)
{
	fprintf(stderr,
		"Usage: usb-atmelprg [-v=vendor] [-d=device] [-v] [options...]\n"
		"   or: usb-uniprog [-v=vendor] [-d=device] [-v] [options...]\n"
		" -v=VVVV   set USB vendor ID (in hex)\n"
		" -d=DDDD   set USB device Id (in hex)\n"
		"        It is more convenient to use the env var FTDI_USB=VVVV:DDDD\n"
		"        instead of using -v and -d. Try \"export FTDI_USB=0403:6002\".\n"
		" -v        increase verbosity (0,1,2)\n"
		" --help    print this\n"
		" --version print version information\n"
		" --edit-ftdi-eeprom (-v, -d default to 0403, 6001)\n"
		"           edit the EEPROM config for the FTDI USB SIO; see docu;\n"
		"           enters interactive mode; will exit after that\n"
		);
	AtmelDevice::DumpSupportedDevices(stderr);
	fprintf(stderr,
		"USB Atmel/Universal Programmer (c) 2005--2009 by Wolfgang Wieser.\n");
	exit(exit_val);
}


static void _PrintVersion()
{
	printf("%s %s (%s)\n","usb-atmelprg",VERSION,VERSDATE);
	printf("%s %s (%s)\n","usb-uniprog",VERSION,VERSDATE);
}


enum CompletionType
{
	CT_None=0,
	CT_File,
	CT_Buffer,
	CT_SetVar,
};

static struct CommandDesc
{
	const char *name;
	CompletionType complete_1,complete_2;
	const char *desc;
} commands[]=
{
	// NOTE: Also add completion handling for new commands. 
	{ NULL,    CT_None, CT_None,
"General commands:" },
	{ "?",     CT_None, CT_None,
		"print help" },
	{ "quit",  CT_None, CT_None,
		"quit the interactive session" },
	{ "reset", CT_None, CT_None,
		"reset communication with USB SIO" },
	{ "ident", CT_None, CT_None,
		"request identification of programmer device (done automatically)" },
	{ "set",   CT_SetVar, CT_None,
		"set communication properties with device; see 'set help'" },
	{ "pwd",   CT_None, CT_None,
		"print current working directory" },
	{ "chdir", CT_File, CT_None,
		"change current working directory" },
	{ "delay", CT_None, CT_None,
		"[NUM]: wait specified number of msec; default 100" },
	{ "avrdevices",CT_None, CT_None,
		"print list of supported AVR microcontrollers" },
	{ "version",CT_None, CT_None,
		"print version information" },
	{ NULL,    CT_None, CT_None,
"Buffer handling commands:" },
	{ "load",  CT_File, CT_None,
		"FILE [BUF]: load specified file into optionally specified buffer" },
	{ "save",  CT_Buffer, CT_File,
		"BUF FILE: save specified buffer in specified file" },
	{ "lsbuf", CT_None, CT_None,
		"list buffers" },
	{ "rmbuf", CT_Buffer, CT_None,
		"BUF: delete buffer; use '*' for all buffers" },
	{ "dump",  CT_Buffer, CT_None,
		"BUF: dump specified buffer" },
	{ "dumpasc",CT_Buffer, CT_None,
		"BUF: dump specified buffer with ASCII column" },
	{ "cmp",   CT_Buffer, CT_Buffer,
		"BUF BUF: compare two buffers" },
	{ NULL,    CT_None, CT_None,
"General IO commands:" },
	{ "dc",    CT_None, CT_None,
		"disconnect from device to be programmed (high-z outputs)" },
	{ "ps",    CT_None, CT_None,
		"query programmer pin status" },
	{ "sup",   CT_None, CT_None,
		"swtich supply: \"on\", \"off\", \"dc\" or value (e.g. \"3.3\"; implies \"on\")" },
	{ NULL,    CT_None, CT_None,
"SPI related commands (for AVR controllers):" },
	{ "connect",CT_None, CT_None,
		"connect programmer to Atmel (done automatically when needed)" },
	{ "erase",  CT_None, CT_None,
		"erase the atmel chip" },
	{ "readflash",CT_Buffer, CT_None,
		"[BUF]: read flash into buffer; default buf name 'flash'" },
	{ "writeflash",CT_Buffer, CT_None,
		"[BUF] [O=offset]: write buffer into flash; default buf name 'flash';\n"
		"offset in words and page aligned ('o' suffix for bytes)" },
	{ "readeeprom",CT_Buffer, CT_None,
		"[BUF]: read EEPROM into buffer; default buf name 'eeprom'" },
	{ "writeeeprom",CT_Buffer, CT_None,
		"[BUF] [O=offset]: write buffer into EEPROM; default buf name 'eeprom';\n"
		"offset in bytes" },
	{ "cmpflash",CT_Buffer, CT_None,
		"[BUF] [O=offset]: read flash and compare with buffer; default buf name 'flash';\n"
		"offset in words ('o' suffix for bytes)" },
	{ "eraseeeprom", CT_None, CT_None,
		"erase EEPROM contents by writing 0xff into all cells (don't do!)" },
	{ "getfl", CT_None, CT_None,
		"read fuse and lock bits" },
	{ "setfl", CT_None, CT_None,
		"set fuse and lock bits; use \"setfl help\" for more information" },
	{ NULL,    CT_None, CT_None,
"JTAG related commands (for CPLDs, and similar; not for AVRs):" },
	{ "jid",  CT_None, CT_None,
		"query and display JTAG IDCODE (one device)" },
	{ "jscanid",CT_None, CT_None,
		"scan all devices for JTAG ID codes" },
	{ "svfprog", CT_File, CT_None,
		"FILE: JTAG download SVF file onto CPLD" },
	{ "jtag", CT_None, CT_None,
		"Enter JTAG mode; use \"quit\" or ^D to leave it again." },
	{ NULL,    CT_None, CT_None,
"Low-level IO (SPI and JTAG; 'p'ull-up not supported by USB-UniProg):" },
	{ "r",     CT_None, CT_None,
		"set reset pin for Atmel: 'l'ow, 'h'high, high-'z', 'p'ull-up" },
	{ "spi",   CT_None, CT_None,
		"set SPI pins for Atmel: 'e'nable, 'p'ullups, high-'z'" },
	{ "d",     CT_None, CT_None,
		"set D-pins 6,7,8 'l'ow, 'h'high, high-'z', 'p'ull-up, e.g. \"d6p78z\"" },
{ NULL, CT_None, CT_None, NULL },
};


static void PrintUsageInteractive()
{
	printf("Interactive commands: (Hit Tab key for completion; "
		"use ';' for command sequence)\n");
	for(int i=0; commands[i].desc; i++)
	{
		if(!commands[i].name)
		{  printf("%s\n",commands[i].desc);  continue;  }
		const char *line=commands[i].desc;
		for(int l=0;;l++)
		{
			const char *eol=strchr(line,'\n');
			if(!eol)  eol=line+strlen(line);
			if(!l)
			{  printf("  %-12s%.*s\n",commands[i].name,eol-line,line);  }
			else
			{  printf("              %.*s\n",eol-line,line);  }
			if(!*eol)  break;
			line=eol+1;
		}
	}
}

static void PrintUsageInteractive_Set()
{
	printf(
		"Command: set [variable] [value]\n"
		"'set' alone without arguments dumps current settings.\n"
		"Host side settings:\n"
		"  verbose NN           verbosity level (0..4)\n"
		"  idle_poll_msec NN    value in msec to poll the USB when not expecing data\n"
		"  busy_poll_usec NN    value in _micro_seconds to poll for USB replies (>=0)\n"
		"  usb_resp_timeout NN  USB response timeout in msec\n"
		"  usb_latency NN       USB SIO packet send latency in msec (1..255)\n"
		"  jtag_delay_fact NN   make JTAG RUNTEST delays longer by a factor 1+NN/10\n");
	
	switch(USB_GetAdapterType())
	{
		case AT_AtmelPrg:
			printf(
				"Settings marked with [AVR only] do not apply to the JTAG mode.\n"
				"Programming device settings:\n"
				"  spi_speed VAL        set IO speed for communication with device to be\n"
				"                       programmed. [AVR only] Possible choices for VAL:\n"
				"                        extreme-> use 0.5625usec SCK HIGH/LOW pulses\n"
				"                        fast   -> use 1.125usec SCK HIGH/LOW pulses\n"
				"                        normal -> use 2.125usec SCK HIGH/LOW pulses (default)\n"
				"                        3,4,.. -> use approx. the specified delay in usec\n"
				"  jtag_speed VAL       JTAG clock speed is 1MHz/(VAL+1)\n"
				"  ext_reset [0/1]      set if reset pin is controlled externaly (1) or not\n"
				"                       (0; default) [AVR only]\n"
				"  try_clock_shift [0/1] try single SCK clock shift during connect [AVR only]\n");
			break;
		case AT_UniProg:
			printf(
				"Settings marked with [AVR only] do not apply to the JTAG mode.\n"
				"Programming device settings:\n"
				"  spi_speed VAL        set IO speed for communication with device to be\n"
				"                       programmed. [AVR only] Possible choices for VAL:\n"
				"                        extreme-> use 0.583usec SCK HIGH/LOW pulses\n"
				"                        fast   -> use 1.083usec SCK HIGH/LOW pulses\n"
				"                        normal -> use 2.167usec SCK HIGH/LOW pulses (default)\n"
				"                        3,4,.. -> LOW/HIGH time in usec\n"
				"  jtag_speed VAL       JTAG clock speed is 12MHz/VAL and 24MHz for VAL=0\n"
				"  io_voltage VAL       IO voltage in V, range 1.5 to 5.0\n"
				"  sup_voltage VAL      Supply voltage in V, range 1.5 to 5.0; 0 to switch off\n"
				"  ext_reset [0/1]      set if reset pin is controlled externaly (1) or not\n"
				"                       (0; default) [AVR only]\n"
				"  try_clock_shift [0/1] try single SCK clock shift during connect [AVR only]\n");
			break;
		default:
			printf(
				"Programming device settings: Cannot be displayed yet since not connected\n"
				"to programmer. Use 'ident' or some similar command.\n");
			break;
	}
}


static char *my_strdup(const char *str)
{
	char *dup=(char*)malloc(strlen(str)+1);
	if(!dup)
	{  fprintf(stderr,"Allocation failure\n");  abort();  }
	strcpy(dup,str);
	return(dup);
}


static void DoSetPrompt()
{
#if HAVE_JTAG_SYSTEM
	if(opmode==MAINSYS && atmel)
	{
		// This is overflow-safe. 
		strncpy(prompt_str,atmel->name,sizeof(prompt_str)-3);
		prompt_str[sizeof(prompt_str)-3]='\0';
		strcat(prompt_str,"> ");
	}
	else if(opmode==JTAGSYS)
	{
		strcpy(prompt_str,"jtag> ");
	}
	else 
#endif
	switch(USB_GetAdapterType())
	{
		case AT_AtmelPrg:
			strcpy(prompt_str,"atmelprg> ");
			break;
		case AT_UniProg:
			strcpy(prompt_str,"uniprog> ");
			break;
		default:
			strcpy(prompt_str,"ready> ");
			break;
	}
	
#if HAVE_LIBREADLINE
	rl_set_prompt(prompt_str);
#endif
}


static char **_atmelprg_readline_completion(const char *text,int start,
	int /*end*/);

#if HAVE_JTAG_SYSTEM
static int EnterJTAGMode()
{
	if(jtag_system_start())  return(1);
	
	opmode=JTAGSYS;
#if HAVE_LIBREADLINE
	rl_attempted_completion_function = NULL;
#endif
	DoSetPrompt();
}

static int LeaveJTAGMode()
{
	jtag_system_cleanup();
	
	opmode=MAINSYS;
#if HAVE_LIBREADLINE
	rl_attempted_completion_function = &_atmelprg_readline_completion;
#endif
	DoSetPrompt();
#if HAVE_LIBREADLINE
	if(got_eof)
	{  rl_forced_update_display();  }  // Needed when leaving via ^D. 
#endif
}
#endif


static int _SetVoltage(int io_voltage,int sup_voltage);

// Connect programmer if not done already. Force an indent if specified. 
static int _DoConnectProgrammer(int force=0)
{
	if(USB_GetAdapterType()!=AT_None && !force)  return(1);
	
	curr_io_voltage_valid=0;
	curr_sup_voltage_valid=0;
	
	USBIdentReply r;
	USB_Status rv=USB_GetBoardIdentification(&r);
	if(rv==USB_UnknownBoard || rv==USB_OK)
	{
		printf(
			"Connected to %s:\n"
			"  Device: %s\n"
			"  Vendor: %s\n"
			"  Firmware version: 0x%x\n"
			"  Protocol version: %u\n"
			,USB_GetIdfString(),
			r.device_name,
			r.device_vendor,
			r.firmware_version,
			r.protocol_version);
		if(verbose)
		{
			const char *atype_str="unknown";
			switch(USB_GetAdapterType())
			{
				case AT_AtmelPrg:  atype_str="AtmelPrg";  break;
				case AT_UniProg:   atype_str="UniProg";   break;
				// default: leave "unknown". 
			}
			printf("  Adapter type: %d (%s)\n",r.atype,atype_str);
		}
	}
	
	if(rv==USB_OK)
	{
		assert(USB_GetAdapterType()!=AT_None);
		if(USB_GetAdapterType()==AT_UniProg)
		{
			// Set default / last IO voltages. 
			// NOTE: curr_*_voltage<0 should never happen!
			_SetVoltage(curr_io_voltage<0 ? 330 : curr_io_voltage,
				curr_sup_voltage<0 ? 330 : curr_sup_voltage);
			// Set default JTAG speed (1MHz). 
			usbatmelprg_jtag_speed=12;
		}
		else if(USB_GetAdapterType()==AT_AtmelPrg)
		{
			// Set default JTAG speed (1MHz). 
			usbatmelprg_jtag_speed=0;
		}
	}
	else
	{
		assert(USB_GetAdapterType()==AT_None);
		if(rv==USB_UnknownBoard)
		{  fprintf(stderr,"Error: unknown board.\n");  }
		else
		{  fprintf(stderr,"Error: %s\n",USB_Status_String(rv));  }
	}
	
	if(USB_GetAdapterType()==AT_None && atmel)
	{  delete atmel;  atmel=NULL;  }
	
	return(USB_GetAdapterType()!=AT_None);
}


static void _DumpProgrammerIOStatus(USBUniProgIOState *s);
static void _DumpProgrammerStatus(_USBProgrammerStatus *s);

// Returns 1 on error, 0 when successful. 
static int _SetResetLine(char state,char do_dump_status)
{
	const char *type_str="???";
	switch(state)
	{
		case 'l':  type_str="LOW";  break;
		case 'h':  type_str="HIGH";  break;
		case 'z':  type_str="high-Z";  break;
		case 'p':  type_str="pull-up";  break;
	}
	
	switch(USB_GetAdapterType())
	{
		case AT_AtmelPrg:
		{
			USBProgrammerStatus stat;
			USB_Status rv=USB_SetPinState(0,state,0,0,0,
				do_dump_status ? &stat : NULL);
			if(rv)
			{
				fprintf(stderr,"Error: Setting reset %s: %s\n",
					type_str,USB_Status_String(rv));
				return(1);
			}
			if(do_dump_status)
			{  _DumpProgrammerStatus(&stat);  }
		}	break;
		case AT_UniProg:
		{
			if(state=='p')
			{  fprintf(stderr,"Pull-ups not supported by USB-UniProg; "
				"falling back to high-Z.\n");  state='z';  }
			
			USBUniProgIOState ios;
			USB_InitUniProgIOState(&ios);
			ios.atmel_reset = state=='p' ? 'z' : state;
			USB_Status rv=USB_SetIOState(&ios);
			if(rv)
			{
				fprintf(stderr,"Error: Setting reset %s: %s\n",
					state=='p' ? "high-Z" : type_str,USB_Status_String(rv));
				return(1);
			}
			if(do_dump_status)
			{  _DumpProgrammerIOStatus(&ios);  }
		}	break;
		default:  assert(0);
	}
	
	return(0);
}


// Set relais; use 0 for no change, +1 for close, -1 for open/release. 
// If ret!=NULL, return values stored there. 
int _SetRelais(int dio,int r5v,int ldo,int gnd,USBUniProgIOState *ret)
{
	if(USB_GetAdapterType()!=AT_UniProg)  return(0);
	
	USBUniProgIOState ios;
	USB_InitUniProgIOState(&ios);
	
	ios.set_relais = 0;
	     if(dio>0)  ios.set_relais |= UAP_RELAIS_DIO_CLOSE;
	else if(dio<0)  ios.set_relais |= UAP_RELAIS_DIO_OPEN;
	     if(r5v>0)  ios.set_relais |= UAP_RELAIS_5V_CLOSE;
	else if(r5v<0)  ios.set_relais |= UAP_RELAIS_5V_OPEN;
	     if(ldo>0)  ios.set_relais |= UAP_RELAIS_LDO_CLOSE;
	else if(ldo<0)  ios.set_relais |= UAP_RELAIS_LDO_OPEN;
	     if(gnd>0)  ios.set_relais |= UAP_RELAIS_GND_CLOSE;
	else if(gnd<0)  ios.set_relais |= UAP_RELAIS_GND_OPEN;
	
	USB_Status rv=USB_SetIOState(&ios);
	if(rv)
	{
		fprintf(stderr,"Error: Setting relais 0x%x: %s\n",
			ios.set_relais,USB_Status_String(rv));
		return(1);
	}
	
	if(ret)
	{  memcpy(ret,&ios,sizeof(USBUniProgIOState));  }
	
	return(0);
}


static void _Voltage2String(char *dest,int len,int volt,int valid);

// io_voltage, sup_voltage: Range 150..500; -1 for no change; 0 to shut down 
// regulator (only sup voltage). 
int _SetVoltage(int io_voltage,int sup_voltage)
{
	char io_volt_str[64];
	char sup_volt_str[64];
	_Voltage2String(io_volt_str,64,io_voltage,1);
	_Voltage2String(sup_volt_str,64,sup_voltage,1);
	printf("Setting voltage levels: IO: %s, supply: %s\n",
		io_voltage>=0 ? io_volt_str : "[unchanged]",
		sup_voltage>=0 ? sup_volt_str : "[unchanged]");
	
	USBUniProgIOState ios;
	USB_InitUniProgIOState(&ios);
	
	if(io_voltage>=0)   ios.io_voltage=io_voltage;
	if(sup_voltage>=0)  ios.sup_voltage=sup_voltage;
	
	USB_Status rv=USB_SetIOState(&ios);
	if(rv)
	{
		fprintf(stderr,"Error: Setting IO/sup voltage %d %d: %s\n",
			io_voltage,sup_voltage,USB_Status_String(rv));
		curr_io_voltage_valid=0;
		curr_sup_voltage_valid=0;
		return(1);
	}
	
	if(io_voltage>=0)
	{  curr_io_voltage=io_voltage;    curr_io_voltage_valid=1;   }
	if(sup_voltage>=0)
	{  curr_sup_voltage=sup_voltage;  curr_sup_voltage_valid=1;  }
	
	return(0);
}



// Connect programmer to atmel if not done already. 
// Force an indent if specified. 
// Return 1 if connected, 0 if not. 
static int _DoConnectAtmel(int force=0)
{
	if(!_DoConnectProgrammer())  return(0);
	
	if(atmel && !force)  return(1);
	if(atmel)
	{  delete atmel;  atmel=NULL;  }
	
	// Electrical connection; Connect grounds. 
	if(_SetRelais(/*dio=*/0,/*r5v=*/0,/*ldo=*/0,/*gnd=*/1,NULL))  return(0);
	
	// If we control the reset pin, then take it LOW first. 
	// Otherwise, it might be floating in HighZ and hence we can never 
	// connect. 
	// Otherwise, set it to High-Z to be sure. 
	if(_SetResetLine(external_reset ? 'z' : 'l',0))  return(0);
	
	// Set the IO voltage and the supply voltage. FIXME!!
	
	// Electrical connection; Connect supply. (FIXME: Do not always do this!)
	if(_SetRelais(/*dio=*/0,/*r5v=*/0,/*ldo=*/1,/*gnd=*/0,NULL))  return(0);
	
	// Electrical connection; Connect DIO. 
	if(_SetRelais(/*dio=*/1,/*r5v=*/0,/*ldo=*/0,/*gnd=*/0,NULL))  return(0);
	
	// Set IO speed for communication with device to be programmed. 
	USB_Status rv=USB_AtmelSetIOSpeed(spi_io_speed);
	if(rv)
	{  fprintf(stderr,"Error setting SPI speed: %s\n",USB_Status_String(rv));
		return(0);  }
	
	rv=USB_AtmelConnect(try_clock_shift);
	if(rv && rv!=USB_ConnectFailed)
	{  fprintf(stderr,"Error during connect: %s\n",USB_Status_String(rv));  }
	if(rv)  return(0);
	
	USBAtmelSignature sig;
	rv=USB_AtmelReadSignature(&sig);
	if(rv)
	{  fprintf(stderr,"Error while reading signature: %s\n",
		USB_Status_String(rv));  return(0);  }
	
	if(verbose)
	{  printf("Signature: vendor=0x%02x, flash_size=0x%02x, device=0x%02x\n",
		sig.vendor,sig.flash_size,sig.device);  }
	
	atmel=GetDeviceBySignature(&sig);
	if(!atmel)
	{
		printf("Unknown signature: "
			"vendor=0x%02x, flash_size=0x%02x, device=0x%02x\n",
			sig.vendor,sig.flash_size,sig.device);
		if(spi_io_speed<=1)
		{  printf("Try slower IO speed using \"set spi_speed normal\".\n");  }
		else
		{
			if(sig.vendor!=0xff || sig.flash_size!=0xff || sig.device!=0xff)
			{  printf("Device not supported on host side. "
				"Please consider implementing it.\n");  }
			if(spi_io_speed<65)
			{  printf("Or, try \"set spi_speed ...\" if device "
				"has slow clock.\n");  }
		}
	}
	else
	{  atmel->DumpIdentification(stdout);  }
	
	return(atmel ? 1 : 0);
}


static void _DoCompare(DataBuffer *dbA,DataBuffer *dbB)
{
	printf("Comparing buffers \"%s\", \"%s\": ",dbA->name,dbB->name);
	fflush(stdout);
	char *paend=dbA->buf+dbA->size;
	char *pbend=dbB->buf+dbB->size;
	size_t difference=0;
	int dump_diff=0;
	for(char *pa=dbA->buf,*pb=dbB->buf; pa<paend && pb<pbend; pa++,pb++)
	{
		if(*pa!=*pb)
		{  ++difference;  }
	}
	if(!difference && dbA->size==dbB->size)
	{  printf("EQUAL (%u bytes)\n",dbA->size);  }
	else if(difference && dbA->size==dbB->size)
	{  printf("%u/%u bytes differ\a\n",difference,dbA->size);  dump_diff=1;  }
	else if(!difference && dbA->size!=dbB->size)
	{  printf("euqal until \"%s\" EOB (%u bytes)\n",
		dbA->size<dbB->size ? dbA->name : dbB->name,
		dbA->size<dbB->size ? dbA->size : dbB->size);  }
	else
	{  printf("%u/%u bytes differ until \"%s\" EOB\a\n",difference,
		dbA->size<dbB->size ? dbA->size : dbB->size,
		dbA->size<dbB->size ? dbA->name : dbB->name);  dump_diff=1;  }
	
	// Write differences: 
	if(dump_diff)
	{
		printf("Differences with byte offset [word offset] page:off delta:\n");
		size_t i=0,last_diff=-1;
		for(char *pa=dbA->buf,*pb=dbB->buf; pa<paend && pb<pbend; pa++,pb++,i++)
		{
			if(*pa==*pb)  continue;
			printf("  0x%04x:  0x%02x <-> 0x%02x  (%c <-> %c)  [0x%04x%c]  "
				"0x%02x:%02x%c %3u\n",
				i,(unsigned int)*(unsigned char*)pa,
				  (unsigned int)*(unsigned char*)pb,
				isprint(*pa) ? *pa : '.',isprint(*pb) ? *pb : '.',
				i/2,(i & 1) ? 'H' : 'L',
				(i/2)>>6,(i/2)&0x3fU,(i & 1) ? 'H' : 'L',
				i-last_diff);
			last_diff=i;
		}
	}
}


static const char *_APinState_String(APinState s)
{
	switch(s)
	{
		case APS_HighZ_0:   return("in:LOW  (Z)");
		case APS_HighZ_1:   return("in:HIGH (Z)");
		case APS_PullUp_0:  return("in:LOW  (P)");
		case APS_PullUp_1:  return("in:HIGH (P)");
		case APS_Out_0:     return("out:LOW");
		case APS_Out_1:     return("out:HIGH");
		default:            return("??? (OOPS)");
	}
}

static void _DumpProgrammerStatus(_USBProgrammerStatus *s)
{
	printf("Programmer status:\n"
		"  SCK: %-12s    MOSI: %-12s    RESET: %-12s\n"
		"  D6:  %-12s    D7:   %-12s    D8:    %-12s\n"
		"  LED: %-12s    MISO: %s\n"
		,_APinState_String(s->sck_state),
			_APinState_String(s->mosi_state),
			_APinState_String(s->reset_state),
		_APinState_String(s->d6_state),
			_APinState_String(s->d7_state),
			_APinState_String(s->d8_state),
		s->led_is_busy_indicator ? "[busy]" : s->led_is_on ? "ON" : "off",
		s->miso_is_high ? "HIGH" : "LOW" );
}


static void _DumpProgrammerIOStatus(USBUniProgIOState *s)
{
	int v[4];
	int i;
	for(i=0; i<4; i++)
	{  v[i]=(s->adc_val[i]+5)/10;  }
	
	printf("Programmer status:\n");
fprintf(stderr,"FIXME: Dump IO status\n");  // FIXME
	
	char relais_str[64];
	if(s->relais_gnd_closed || s->relais_dio_closed || s->relais_5v_closed || 
		s->relais_ldo_closed)
	{
		snprintf(relais_str,64," closed:%s%s%s%s"
			,s->relais_gnd_closed ? " GND" : "",
			s->relais_dio_closed ? " IO" : "",
			s->relais_5v_closed ? " 5V" : "",
			s->relais_ldo_closed ? " LDO" : "");
	}
	else
	{  strcpy(relais_str,": all open");  }
	
	if(verbose)
	{  printf("  Relais%-30s     LED: %s\n",
		relais_str,
		s->busy_led_enabled ? "[busy]" : s->led_on ? "on" : "off");  }
	else
	{  printf("  Relais%s\n",relais_str);  }
	
	printf("  ADC: Supplies: %d.%02dV, %d.%02dV;  IO: %d.%02dV;  LDO: %d.%02dV\n"
		,v[0]/100,v[0]%100,
		v[1]/100,v[1]%100,
		v[2]/100,v[2]%100,
		v[3]/100,v[3]%100);
}


static void _AtmelDisconnect()
{
	if(USB_GetAdapterType()==AT_None)  return;
	
	switch(USB_GetAdapterType())
	{
		case AT_AtmelPrg:
		{
			USBProgrammerStatus stat;
			USB_Status rv=USB_SetPinState('p',external_reset ? 'z' : 'p',
				'z','z','z',&stat);
			if(rv)
			{  fprintf(stderr,"Error: %s\n",USB_Status_String(rv));  }
			else
			{  _DumpProgrammerStatus(&stat);  }
		} break;
		case AT_UniProg:
		{
			USBUniProgIOState ios;
			USB_InitUniProgIOState(&ios);
			//ios->atmel_reset = external_reset ? 'z' : 'z';  // 'h'
			// Do all via ctrl_mask. 
			ios.ctrl_mask = UNIPROG_BIT_DDR_TCK_SCK | UNIPROG_BIT_DDR_TDO_MISO | 
				UNIPROG_BIT_DDR_TDI_MOSI | UNIPROG_BIT_DDR_TMS_RESET;
			ios.ddr_bits = 0x00;  // input -> high-Z. 
			
			ios.set_relais = UAP_RELAIS_DIO_OPEN;
			
			USB_Status rv=USB_SetIOState(&ios);
			if(rv)
			{  fprintf(stderr,"Error: %s\n",USB_Status_String(rv));  }
// FIXME: Disconnect relais etc. 
// FIXME Do this in correct sequence!
			
// FIXME: Currently, dc does not disconnect power lines. Do we want that?
			
			if(!_SetRelais(/*dio=*/-1,/*r5v=*/-1,/*ldo=*/0,/*gnd=*/0,&ios))
			{  _DumpProgrammerIOStatus(&ios);  }
		} break;
		default:  assert(0);
	}
}

static void _Voltage2String(char *dest,int len,int volt,int valid)
{
	char v_tmp[32];
	if(volt>0)
	{  snprintf(v_tmp,32,"%d.%02d V",volt/100,volt%100);  }
	else if(!volt)
	{  strcpy(v_tmp,"off");  }
	else
	{  strcpy(v_tmp,"???");  }
	
	snprintf(dest,len,"%s%s",
		v_tmp,(valid || volt<0) ? "" : "  (possibly invalid!)");
}

static void _DumpSettings()
{
	printf(
		"Host side settings:\n"
		"  verbose:          %d\n"
		"  idle_poll_msec:   %d msec\n"
		"  busy_poll_usec:   %d usec\n"
		"  usb_resp_timeout: %d msec\n"
		"  usb_latency:      %d msec\n"
		"  jtag_delay_fact:  %d\n"
		"Programming device settings:\n"
		,verbose,
		idle_poll_msec,
		busy_poll_usec,
		usb_resp_timeout,
		USB_GetLatency(),
		jtag_delay_prolong_fact);
	switch(USB_GetAdapterType())
	{
		case AT_AtmelPrg:
		{
			char spi_speed_str[24];
			switch(spi_io_speed)
			{
				case 0:  strcpy(spi_speed_str,"extreme (0.5625 usec)"); break;
				case 1:  strcpy(spi_speed_str,"fast (1.125 usec)");     break;
				case 2:  strcpy(spi_speed_str,"normal (2.125 usec)");   break;
				default:
					snprintf(spi_speed_str,24,"%d usec delay",spi_io_speed);
					break;
			}
			
			int jspeed10 = (10000+(usbatmelprg_jtag_speed+1)/2)/
				((int)usbatmelprg_jtag_speed+1);
			printf(
				"  spi_speed:        %s\n"
				"  jtag_speed:       %d (%d.%d kHz)\n"
				"  ext_reset:        %d\n"
				"  try_clock_shift:  %d\n"
				,spi_speed_str,
				(int)usbatmelprg_jtag_speed,jspeed10/10,jspeed10%10,
				external_reset,
				(int)try_clock_shift);
		}	break;
		case AT_UniProg:
		{
			char spi_speed_str[24];
			switch(spi_io_speed)
			{
				case 0:  strcpy(spi_speed_str,"extreme (0.583 usec)"); break;
				case 1:  strcpy(spi_speed_str,"fast (1.083 usec)");     break;
				case 2:  strcpy(spi_speed_str,"normal (2.167 usec)");   break;
				default:
					snprintf(spi_speed_str,24,"%d usec delay",spi_io_speed);
					break;
			}
			
			char io_volt_str[64];
			char sup_volt_str[64];
			_Voltage2String(io_volt_str,64,curr_io_voltage,
				curr_io_voltage_valid);
			_Voltage2String(sup_volt_str,64,curr_sup_voltage,
				curr_io_voltage_valid);
			
			int jtag_kHz = usbatmelprg_jtag_speed ? 
				12000/usbatmelprg_jtag_speed : 24000;
			printf(
				"  spi_speed:        %s\n"
				"  jtag_speed:       %d (%d kHZ)\n"
				"  io_voltage:       %s\n"
				"  sup_voltage:      %s\n"
				"  ext_reset:        %d\n"
				"  try_clock_shift:  %d\n"
				,spi_speed_str,
				(int)usbatmelprg_jtag_speed,jtag_kHz,
				io_volt_str,
				sup_volt_str,
				external_reset,
				(int)try_clock_shift);
		}	break;
		default:
			printf("  [not yet connected; use 'ident' or some similar command]\n");
			break;
	}
}


static void _BitCharToBitMask(USBUniProgIOState *ios,char c,uint8 mask)
{
	if(!c)  return;
	
	ios->ctrl_mask |= mask;
	
	switch(c)
	{
		case 'l':  ios->ddr_bits|=mask;  ios->level_bits&=~mask;  break;
		case 'h':  ios->ddr_bits|=mask;  ios->level_bits|=mask;   break;
		case 'z':  ios->ddr_bits&=~mask;  break;
		default:  assert(0);
	}
}


// Single command: without semicolon inside. 
static void ExecuteSingleCommand(char *cmd_buf)
{
#if HAVE_JTAG_SYSTEM
	if(opmode==JTAGSYS)
	{
		int rv=jtag_system_parse_line(cmd_buf);
		// rv=0 for exit. 
		if(!rv)
		{  LeaveJTAGMode();  }
		return;
	}
#endif
	
	char *cmd=cmd_buf;
	char *end=cmd;
	while(!isspace(*end) && *end)  ++end;
	if(*end) *end++='\0';
	
	const int max_args=32;
	char *args[max_args];
	for(int i=0; i<max_args; i++)  args[i]=NULL;
	int nargs=0;
	for(;;)
	{
		while(isspace(*end))  ++end;
		if(!*end)  break;
		if(nargs>=max_args)
		{  fprintf(stderr,"Too many arguments.\n");  break;  }
		args[nargs++]=end;
		while(!isspace(*end) && *end)  ++end;
		if(!*end)  break;
		*end++='\0';
	}
	
	if(verbose>2)
	{  fprintf(stderr,"cmd: \"%s\"; args: %s %s %s ...\n",
		cmd_buf,args[0],args[1],args[2]);  }
	
	if(!strcmp(cmd,"?") || !strcmp(cmd,"help"))
	{
		if(args[0] && !strcmp(args[0],"set"))
		{  PrintUsageInteractive_Set();  }
		else
		{  PrintUsageInteractive();  }
	}
	else if(!strcmp(cmd,"version"))
	{
		_PrintVersion();
	}
	else if(!strcmp(cmd,"avrdevices"))
	{
		AtmelDevice::DumpSupportedDevices(stdout);
	}
	else if(!strcmp(cmd,"quit"))
	{
		++got_eof;
#if HAVE_LIBREADLINE
		rl_set_prompt("");
#endif
		return;
	}
	else if(!strcmp(cmd,"reset"))
	{
		USB_Reset();
	}
	else if(!strcmp(cmd,"set"))
	{
		if(!args[0])
		{  _DumpSettings();  }
		else if(!strcmp(args[0],"help"))
		{  PrintUsageInteractive_Set();  }
		else if(!strcmp(args[0],"ext_reset"))
		{
			if(!args[1])  goto reqvalmissing;
			external_reset=strtol(args[1],NULL,0);
			external_reset = external_reset ? 1 : 0;
		}
		else if(!strcmp(args[0],"spi_speed"))
		{
			if(!args[1])  goto reqvalmissing;
			if(!strncmp(args[1],"normal",strlen(args[1])))
			{  spi_io_speed=2;  }
			else if(!strncmp(args[1],"fast",strlen(args[1])))
			{  spi_io_speed=1;  }
			else if(!strncmp(args[1],"extreme",strlen(args[1])))
			{  spi_io_speed=0;  }
			else
			{
				spi_io_speed=strtol(args[1],NULL,0);
				if(spi_io_speed<3)  spi_io_speed=3;
				if(spi_io_speed>=0xffU)  spi_io_speed=0xfeU;  // 0xff reserved.
			}
			// If we're connected, force a re-connect. 
			if(atmel)
			{  _DoConnectAtmel(/*force=*/1);  }
		}
		else if(!strcmp(args[0],"jtag_speed"))
		{
			if(!args[1])  goto reqvalmissing;
			int tmp=strtol(args[1],NULL,0);
			usbatmelprg_jtag_speed=tmp<0 ? 0U : tmp>255 ? 255U : (uint8)tmp;
		}
		else if(!strcmp(args[0],"idle_poll_msec"))
		{
			if(!args[1])  goto reqvalmissing;
			idle_poll_msec=strtol(args[1],NULL,0);
			idle_poll_msec = idle_poll_msec<=0 ? 1 : idle_poll_msec;
		}
		else if(!strcmp(args[0],"busy_poll_usec"))
		{
			if(!args[1])  goto reqvalmissing;
			busy_poll_usec=strtol(args[1],NULL,0);
			busy_poll_usec = busy_poll_usec<0 ? 0 : busy_poll_usec;
		}
		else if(!strcmp(args[0],"usb_resp_timeout"))
		{
			if(!args[1])  goto reqvalmissing;
			usb_resp_timeout=strtol(args[1],NULL,0);
			usb_resp_timeout = usb_resp_timeout<=0 ? 1 : usb_resp_timeout;
		}
		else if(!strcmp(args[0],"jtag_delay_fact"))
		{
			if(!args[1])  goto reqvalmissing;
			jtag_delay_prolong_fact=strtol(args[1],NULL,0);
			jtag_delay_prolong_fact = jtag_delay_prolong_fact<-10 ? 
				-10 : jtag_delay_prolong_fact;
		}
		else if(!strcmp(args[0],"verbose"))
		{
			if(!args[1])  goto reqvalmissing;
			verbose=strtol(args[1],NULL,0);
			verbose=verbose<0 ? 0 : verbose;
		}
		else if(!strcmp(args[0],"try_clock_shift"))
		{
			if(!args[1])  goto reqvalmissing;
			try_clock_shift=strtol(args[1],NULL,0);
			try_clock_shift = try_clock_shift ? 1 : 0;
		}
		else if(!strcmp(args[0],"usb_latency"))
		{
			if(!args[1])  goto reqvalmissing;
			int latency=strtol(args[1],NULL,0);
			USB_SetLatency(latency);
		}
		else if(USB_GetAdapterType()==AT_UniProg && 
			(!strcmp(args[0],"io_voltage") || !strcmp(args[0],"sup_voltage")))
		{
			if(!args[1])  goto reqvalmissing;
			double val = strtod(args[1],NULL);
			int ival = (int)(val*100+0.5);
			int force = args[2] && args[2][0]=='!';
			if((ival>500 || ival<150) && ival!=0)
			{  fprintf(stderr,"Voltage specification out of range 1.5 to 5V.\n");  }
			else if(!force && args[0][0]=='i' && curr_sup_voltage_valid && 
				curr_sup_voltage>0 && curr_sup_voltage<ival)
			{  fprintf(stderr,"Attempt to make IO voltage larger than supply voltage.\n"
				"  Use '!' to force this change.\n");  }
			else if(!force && args[0][0]=='s' && curr_io_voltage_valid && 
				ival>0 && curr_io_voltage>ival)
			{  fprintf(stderr,"Attempt to make supply voltage smaller than IO voltage.\n"
				"  Use '!' to force this change.\n");  }
			else
			{
				_SetVoltage(args[0][0]=='i' ? ival : -1,
					args[0][0]=='s' ? ival : -1);
			}
		}
		else
		{  fprintf(stderr,"Unknown variable '%s'%s. Try 'set help'.\n",
			args[0],USB_GetAdapterType()==AT_None ? 
				" or currently not connected" : "");  }
		
		if(0) { reqvalmissing: fprintf(stderr,"Required value missing.\n"); }
	}
	else if(!strcmp(cmd,"ident"))
	{
		_DoConnectProgrammer(/*force=*/1);
	}
	else if(!strcmp(cmd,"r") || !strcmp(cmd,"rl") || !strcmp(cmd,"rh") || 
		!strcmp(cmd,"rz") || !strcmp(cmd,"rp"))
	{
		char how=cmd[1];
		if(!how && args[0])
		{  how=args[0][0];  }
		if(!how || (how!='l' && how!='h' && how!='z' && how!='p'))
		{  fprintf(stderr,"Missing state: 'l'ow, 'h'high, high-'z', "
			"'p'ull-ups\n");  }
		else if(external_reset && how!='z' && how!='p')
		{  fprintf(stderr,"Cannot set reset; is externally controlled "
			"(see 'set help')\n");  }
		else if(_DoConnectProgrammer())
		{  _SetResetLine(how,1);  }
	}
	else if(!strcmp(cmd,"spi"))
	{
		char how=args[0][0];
		if(_DoConnectProgrammer())
		{
			switch(USB_GetAdapterType())
			{
				case AT_AtmelPrg:
				{
					if(!how || (how!='e' && how!='p' && how!='z'))
					{  fprintf(stderr,"Missing state: 'e'nable, 'p'ullups, "
						"high-'z'\n");  }
					else
					{
						USBProgrammerStatus stat;
						USB_Status rv=USB_SetPinState(how,0,0,0,0,&stat);
						if(rv)
						{ fprintf(stderr,"Error: %s\n",USB_Status_String(rv)); }
						else
						{  _DumpProgrammerStatus(&stat);  }
					}
				} break;
				case AT_UniProg:
				{
					if(how=='p')
					{  fprintf(stderr,"Pull-ups not supported by USB-UniProg; "
						"falling back to high-Z.\n");  how='z';  }
					if(!how || (how!='e' && how!='z'))
					{  fprintf(stderr,"Missing state: 'e'nable, high-'z'\n");  }
					else
					{
						USBUniProgIOState ios;
						USB_InitUniProgIOState(&ios);
						if(how=='e')
						{
							ios.ctrl_mask = UNIPROG_BIT_DDR_TCK_SCK | UNIPROG_BIT_DDR_TDI_MOSI;
							ios.ddr_bits = UNIPROG_BIT_DDR_TCK_SCK | UNIPROG_BIT_DDR_TDI_MOSI;
						}
						else
						{
							ios.ctrl_mask = UNIPROG_BIT_DDR_TCK_SCK | UNIPROG_BIT_DDR_TDI_MOSI;
							ios.ddr_bits = 0;
						}
						ios.level_bits=0x00;
						
						USB_Status rv=USB_SetIOState(&ios);
						if(rv)
						{ fprintf(stderr,"Error: %s\n",USB_Status_String(rv)); }
						else
						{  _DumpProgrammerIOStatus(&ios);  }
					}
				} break;
				default:  assert(0);
			}
		}
	}
	else if(!strcmp(cmd,"dc"))
	{
		if(_DoConnectProgrammer())
		{
			_AtmelDisconnect();
			
			if(atmel)
			{  delete atmel; atmel=NULL;  }
		}
	}
	else if(!strcmp(cmd,"ps"))
	{
		if(_DoConnectProgrammer())
		{
			switch(USB_GetAdapterType())
			{
				case AT_AtmelPrg:
				{
					USBProgrammerStatus stat;
					USB_Status rv=USB_SetPinState(0,0,0,0,0,&stat);
					if(rv)
					{  fprintf(stderr,"Error: %s\n",USB_Status_String(rv));  }
					else
					{  _DumpProgrammerStatus(&stat);  }
				} break;
				case AT_UniProg:
				{
					USBUniProgIOState ios;
					USB_InitUniProgIOState(&ios);
					USB_Status rv=USB_SetIOState(&ios);
					if(rv)
					{  fprintf(stderr,"Error: %s\n",USB_Status_String(rv));  }
					else
					{  _DumpProgrammerIOStatus(&ios);  }
				} break;
				default:  assert(0);
			}
		}
	}
	else if(!strcmp(cmd,"sup"))
	{
		// FIXME: Merge functionality with "set sup_voltage"!!!
		//         Allow something like "sup 5 ; connect 3.3 ; ... ; dc ; sup off"
		
		// Inspect arguments. 
		if(!args[0])  goto reqvalmissing;
		
		// "on", "off": switch the relais. 
		// "0" -> shutdown; "1.5"..."5" -> tune voltage
		int on_off=0;
		int volts=-1;
		if(!strcmp(args[0],"on"))  on_off=+1;
		else if(!strcmp(args[0],"off"))  on_off=-1;
		else if(!strcmp(args[0],"dc"))  on_off=-2;
		else
		{
			double val = strtod(args[0],NULL);
			volts = (int)(val*100+0.5);
			if((volts>500 || volts<150) && volts!=0)
			{  fprintf(stderr,"Voltage specification out of range 1.5 to 5V.\n");  }
			// Positive value implies "on". 
			if(volts>0)
			{  on_off=+1;  }
		}
		
		// Make one try to connect if not connected. 
		if(USB_GetAdapterType()==AT_None)
		{
			// Not connected; we should set the right voltage right in the 
			// beginning. 
			if(volts>=0)
			{
				curr_sup_voltage=volts;
				if(curr_io_voltage>volts && volts)
				{  curr_io_voltage=volts;  }
			}
			
			_DoConnectProgrammer();
		}
		
		if(USB_GetAdapterType()==AT_UniProg)
		{
			// Alright: Set the voltage and so on. 
			if(volts>=0)
			{  _SetVoltage(-1,volts);  }
			
			if(on_off>0)  // on
			{
				// FIXME: Programmer should close GND before LDO. 
				//        programmer shoudl soft-start LDO. 
				_SetRelais(/*dio=*/0,/*r5v=*/0,/*ldo=*/+1,/*gnd=*/+1,NULL);
			}
			else if(on_off<0)  // off or disconnect
			{
				// Disconnect LDO. 
				_SetRelais(/*dio=*/0,/*r5v=*/0,/*ldo=*/-1,/*gnd=*/0,NULL);
				// Also disconnect GND if requested. 
				if(on_off==-2)  // off disconnect
				{  _SetRelais(/*dio=*/0,/*r5v=*/0,/*ldo=*/-1,/*gnd=*/-1,NULL); }
			}
		}
		else 
		{
			fprintf(stderr,"Not connected to programmer or \"sup\" "
				"command not supported\n");
		}
	}
	else if(!strcmp(cmd,"load"))
	{
		char *file_name=args[0];
		char *buf_name=args[1];
		if(!file_name)
		{  fprintf(stderr,"Missing filename.\n");  }
		else
		{
			// Let buffer name default to file name without path: 
			if(!buf_name)
			{
				buf_name=strrchr(file_name,'/');
				if(buf_name)  ++buf_name;
				else buf_name=file_name;
			}
			
			// Read in buffer. 
			DataBuffer *db=buf_list.GetDataBuffer(buf_name,/*do_create=*/1);
			assert(db);
			db->ReadBinaryFile(file_name);
		}
	}
	else if(!strcmp(cmd,"save"))
	{
		char *buf_name=args[0];
		char *file_name=args[1];
		if(!buf_name || !file_name)
		{  fprintf(stderr,"Missing buffer or file name.\n");  }
		else
		{
			DataBuffer *db=buf_list.GetDataBuffer(buf_name,/*do_create=*/0);
			if(!db)
			{  fprintf(stderr,"No such buffer -- %s\n",buf_name);  }
			else
			{  db->WriteBinaryFile(file_name);  }
		}
	}
	else if(!strcmp(cmd,"dump") || !strcmp(cmd,"dumpasc"))
	{
		int with_ascii=!strcmp(cmd,"dumpasc");
		char *buf_name=args[0];
		if(!buf_name)
		{  fprintf(stderr,"Missing buffer name.\n");  }
		else
		{
			DataBuffer *db=buf_list.GetDataBuffer(buf_name,/*do_create=*/0);
			if(!db)
			{  fprintf(stderr,"No such buffer -- %s\n",buf_name);  }
			else
			{  db->HexDumpBuffer(stdout,with_ascii);  }
		}
	}
	else if(!strcmp(cmd,"cmp"))
	{
		char *bufA=args[0];
		char *bufB=args[1];
		if(!bufA || !bufB)
		{  fprintf(stderr,"Compare requires 2 buffer names.\n");  }
		else
		{
			DataBuffer *dbA=buf_list.GetDataBuffer(bufA,/*do_create=*/0);
			DataBuffer *dbB=buf_list.GetDataBuffer(bufB,/*do_create=*/0);
			if(!dbA || !dbB)
			{  fprintf(stderr,"-No such buffer -- %s\n",!dbA ? bufA : bufB);  }
			else
			{  _DoCompare(dbA,dbB);  }
		}
	}
	else if(!strcmp(cmd,"lsbuf"))
	{
		size_t sum_bytes=0;
		int nbuffers=0;
		for(DataBuffer *i=buf_list.head; i; i=i->next)
		{
			printf("  %6d  %s\n",i->size,i->name);
			sum_bytes+=i->size;
			++nbuffers;
		}
		printf("Total: %u bytes in %d buffers\n",sum_bytes,nbuffers);
	}
	else if(!strcmp(cmd,"rmbuf"))
	{
		if(!args[0])
		{  fprintf(stderr,"Missing buffer name.\n");  }
		else if(!strcmp(args[0],"*"))
		{
			// Delete all buffers. 
			buf_list.Clear();
		}
		else
		{
			if(buf_list.RemoveDataBuffer(args[0]))
			{  fprintf(stderr,"No such buffer -- %s\n",args[0]);  }
		}
	}
	else if(!strcmp(cmd,"connect"))
	{
		_DoConnectAtmel(/*force=*/1);
	}
	else if(!strcmp(cmd,"readflash") || !strcmp(cmd,"readeeprom"))
	{
		int is_flash = !strcmp(cmd,"readflash");
		const char *buf_name = args[0] ? args[0] : 
			(is_flash ? "flash" : "eeprom");
		if(!buf_name)  // Cannot happen, currently. 
		{  fprintf(stderr,"Missing buffer name.\n");  }
		else
		{
			DataBuffer *db=buf_list.GetDataBuffer(buf_name,/*do_create=*/1);
			if(_DoConnectAtmel())
			{
				if(is_flash)  atmel->ReadFlash(db);
				else  atmel->ReadEEPROM(db);
			}
		}
	}
	else if(!strcmp(cmd,"cmpflash") || !strcmp(cmd,"cmpeeprom"))
	{
		int is_flash = !strcmp(cmd,"cmpflash");
		const char *buf_name = args[0] ? args[0] : 
			(is_flash ? "flash" : "eeprom");
		if(!buf_name)  // Cannot happen, currently. 
		{  fprintf(stderr,"Missing buffer name.\n");  }
		else
		{
			DataBuffer *db=buf_list.GetDataBuffer(buf_name,/*do_create=*/0);
			if(!db)
			{  fprintf(stderr,"No such buffer -- %s\n",buf_name);  }
			else
			{
				int offset=0;
				if(args[1] && !strncmp(args[1],"O=",2))
				{
					char *endptr=NULL;
					offset=strtol(args[1]+2,&endptr,0);
					if(endptr && *endptr=='o')
					{  offset/=2;  }
				}
				if(_DoConnectAtmel())
				{
					DataBuffer *tmp = new DataBuffer("tmp");  // (not in list)
					if(is_flash)
					{  atmel->ReadFlash(tmp,offset,
						MIN<size_t>(db->size+1,atmel->flash_size*2)/2);  }
					else
					{  atmel->ReadEEPROM(tmp,offset,
						MIN<size_t>(db->size,atmel->eeprom_size));  }
					_DoCompare(db,tmp);
					delete tmp;
				}
			}
		}
	}
	else if(!strcmp(cmd,"writeflash") || !strcmp(cmd,"writeeeprom"))
	{
		int is_flash = !strcmp(cmd,"writeflash");
		const char *buf_name = args[0] ? args[0] : 
			(is_flash ? "flash" : "eeprom");
		// FIXME: We should actually perform somewhat more intelligent argument 
		//        handling and add a skip=NNN as well. 
		if(!buf_name)  // Cannot happen, currently. 
		{  fprintf(stderr,"Missing buffer name.\n");  }
		else
		{
			DataBuffer *db=buf_list.GetDataBuffer(buf_name,/*do_create=*/0);
			if(!db)
			{  fprintf(stderr,"No such buffer -- %s\n",buf_name);  }
			else
			{
				int offset=0;
				if(args[1] && !strncmp(args[1],"O=",2))
				{
					char *endptr=NULL;
					offset=strtol(args[1]+2,&endptr,0);
					if(endptr && *endptr=='o')
					{  offset/=2;  }
				}
				if(_DoConnectAtmel())
				{
					if(is_flash)  atmel->WriteFlash(db,offset);
					else  atmel->WriteEEPROM(db,offset);
				}
			}
		}
	}
	else if(!strcmp(cmd,"eraseeeprom"))
	{
		if(_DoConnectAtmel())
		{  atmel->EraseEEPROM();  }
	}
	else if(!strcmp(cmd,"erase"))
	{
		if(_DoConnectAtmel())
		{
			if(atmel->EraseAll()==-1)
			{
				// Special case: -1 means disconnect. 
				_AtmelDisconnect();
				if(atmel)
				{  delete atmel; atmel=NULL;  }
			}
		}
	}
	else if(!strcmp(cmd,"getfl"))
	{
		if(_DoConnectAtmel())
		{
			atmel->DumpFuseLock(stdout);
		}
	}
	else if(!strcmp(cmd,"setfl"))
	{
		if(_DoConnectAtmel())
		{
			atmel->WriteFuseLock(stdout,args,nargs);
		}
	}
	else if(!strcmp(cmd,"pwd"))
	{
		// Okay, so if somebody is using a longer path, (s)he needs to 
		// code something more fancy here...
		char buf[1024];
		getcwd(buf,sizeof(buf));
		printf("%s\n",buf);
		fflush(stdout);
	}
	else if(!strcmp(cmd,"chdir"))
	{
		char *dir=args[0];
		if(!dir)
		{  fprintf(stderr,"Missing directory.\n");  }
		else
		{
			int rv=chdir(dir);
			if(rv<0)
			{  fprintf(stderr,"Failed to chdir to \"%s\": %s\n",
				dir,strerror(errno));  }
		}
	}
	else if(!strcmp(cmd,"jid"))
	{
		if(_DoConnectProgrammer())
		{  JTagTAP::QueryIDCode();  }
	}
	else if(!strcmp(cmd,"jscanid"))
	{
		if(_DoConnectProgrammer())
		{  JTagTAP::ScanIDCodes();  }
	}
	else if(!strcmp(cmd,"svfprog"))
	{
		char *file=args[0];
		if(!file)
		{  fprintf(stderr,"Missing SF file.\n");  }
		else if(_DoConnectProgrammer())
		{
			JTagTAP::SVFProgramFile(file);
		}
	}
	else if(!strcmp(cmd,"delay"))
	{
		if(args[0] && (!strcmp(args[0],"k") || !strcmp(args[0],"key")))
		{
			printf("Press return to continue...");
			// FIXME: This is a temporary hack which does not work properly 
			//        with readline support enabled. 
			char tmp[3];
			fgets(tmp,2,stdin);
		}
		else
		{
			int tmp = args[0] ? strtol(args[0],NULL,0) : 100;
			usleep(tmp*1000);  // Can be interrupted. 
		}
	}
#if HAVE_JTAG_SYSTEM
	else if(!strcmp(cmd,"jtag"))
	{
		EnterJTAGMode();
	}
#endif
	// NOTE: This one must be last!
	else 
	{
		// Try to parse a valid "d" command. They may be run-together or 
		// spaced apart and contain more than one spec. 
		// "d6z" or "d 6z" or "d 6 z 7 p" or "d6z7h" or "d6p78z"...
		bool is_valid_d=0;
		if(cmd[0]=='d')
		{
			is_valid_d=1;
			const char *p=&cmd[1];
			char bit_sel[3]={0,0,0};
			char bit_val[3]={0,0,0};
			int aidx=0;
			for(;;)
			{
				while(*p)
				{
					char pp=*p;
					switch(pp)
					{
						case '6':  bit_sel[0]=1;  break;
						case '7':  bit_sel[1]=1;  break;
						case '8':  bit_sel[2]=1;  break;
						case 'l':
						case 'h':
						case 'p':
						case 'z':
						{
							if(pp=='p' && USB_GetAdapterType()==AT_UniProg)
							{
								fprintf(stderr,"Pull-ups not supported by "
									"USB-UniProg; falling back to high-Z.\n");
								pp='z';
							}
							int ns=0;
							if(bit_sel[0]) { bit_val[0]=pp; ++ns; }
							if(bit_sel[1]) { bit_val[1]=pp; ++ns; }
							if(bit_sel[2]) { bit_val[2]=pp; ++ns; }
							bit_sel[0]=bit_sel[1]=bit_sel[2]=0;
							if(!ns)
							{  is_valid_d=0;  goto leave_d;  }
						}	break;
						default:  is_valid_d=0;  goto leave_d;
					}
					++p;
				}
				if(aidx>=nargs) break;
				p=args[aidx++];
			}
			
			if(_DoConnectProgrammer())
			{
				switch(USB_GetAdapterType())
				{
					case AT_AtmelPrg:
					{
						USBProgrammerStatus stat;
						USB_Status rv=USB_SetPinState(0,0,bit_val[0],
							bit_val[1],bit_val[2],&stat);
						if(rv)
						{  fprintf(stderr,"Error: %s\n",
							USB_Status_String(rv));  }
						else
						{  _DumpProgrammerStatus(&stat);  }
					} break;
					case AT_UniProg:
					{
						USBUniProgIOState ios;
						USB_InitUniProgIOState(&ios);
						ios.ctrl_mask = 0;
						ios.ddr_bits = 0;
						ios.level_bits = 0;
						_BitCharToBitMask(&ios,bit_val[0],UNIPROG_BIT_DDR_D6);
						_BitCharToBitMask(&ios,bit_val[1],UNIPROG_BIT_DDR_D7);
						_BitCharToBitMask(&ios,bit_val[2],UNIPROG_BIT_DDR_D8RES);
						USB_Status rv=USB_SetIOState(&ios);
						if(rv)
						{  fprintf(stderr,"Error: %s\n",USB_Status_String(rv));  }
						else
						{  _DumpProgrammerIOStatus(&ios);  }
					} break;
					default:  assert(0);
				}
			}
			
			leave_d:;
			if(!is_valid_d)
			{
				fprintf(stderr,"Unknown command '%s' or illegal 'd' command. "
					"Try '?'\n",cmd);
			}
		}
		else if(strlen(cmd))
		{
			fprintf(stderr,"Unknown command '%s'. Try '?'\n",cmd);
		}
	}
}

// Command may include semicolon. 
static void ExecuteCommand(char *cmd_buf)
{
	// Split command into single commands at ";". 
	char *cmd_start=cmd_buf;
	bool last=0;
	for(;;)
	{
		while(isspace(*cmd_start))  ++cmd_start;
		char *cmd_end=cmd_start;
		while(*cmd_end && *cmd_end!=';')  ++cmd_end;
		if(*cmd_end==';')
		{  *cmd_end='\0';  }
		else if(!*cmd_end)
		{  last=1;  }
		ExecuteSingleCommand(cmd_start);
		if(last)  break;
		cmd_start=cmd_end+1;
		
		if(caught_sigint)
		{
			fprintf(stderr,"Command sequence terminated by SIGINT.\n");
			break;
		}
	}
	
	DoSetPrompt();
}


#if HAVE_LIBREADLINE
static void _readline_callback(char *cmd_buf)
{
	if(!cmd_buf)
	{
		printf("^D\n");
		++got_eof;
		rl_set_prompt("");
		return;
	}
	
	do {
		char *cmd=cmd_buf;
		if(*cmd=='\0')  break;
		
		while(isspace(*cmd))  ++cmd;
		
		// Do not memorize commands beginning with a space. 
		if(cmd==cmd_buf)			
		{
			// Do not memorize commands which are the same as the 
			// previous command. 
// This is commented out because it does not what I want. 
//int where=where_history();
//HIST_ENTRY *he=history_get(where);
//if(!he || strcmp(cmd,he->line))
//			{  add_history(cmd);  }
			static char *prev_cmd=NULL;
			if(!prev_cmd || strcmp(prev_cmd,cmd))
			{  add_history(cmd);  }
			if(prev_cmd)  free(prev_cmd);
			prev_cmd=my_strdup(cmd);
		}
		
		char *buf_end=cmd+strlen(cmd);
		while(isspace(*--buf_end))  *buf_end='\0';
		
		ExecuteCommand(cmd);
	} while(0);
	
	free(cmd_buf);
}


// Generator function for command completion.  STATE lets us
// know whether to start from scratch; without any state
// (i.e. STATE == 0), then we start at the top of the list. */
static char *_atmelprg_command_generator(const char *text,int state)
{
	static int list_index, len;
	
	/* If this is a new word to complete, initialize now.  This
	   includes saving the length of TEXT for efficiency, and
	   initializing the index variable to 0. */
	if(!state)
	{
		list_index = 0;
		len = strlen(text);
	}
	
	/* Return the next name which partially matches from the
	   command list. */
	for(;;)
	{
		if(!commands[list_index].desc)  break;
		const char *name=commands[list_index++].name;
		if(!name)  continue;
		
		if(!strncmp(name,text,len))
		{  return(my_strdup(name));  }
	}
	
	/* If no names matched, then return NULL. */
	rl_attempted_completion_over=1;
	return(NULL);
}

static char *_atmelprg_buffer_generator(const char *text,int state)
{
	static int list_index, len;
	
	/* If this is a new word to complete, initialize now.  This
	   includes saving the length of TEXT for efficiency, and
	   initializing the index variable to 0. */
	if(!state)
	{
		list_index = 0;
		len = strlen(text);
	}
	
	/* Return the next name which partially matches from the
	   command list. */
	DataBuffer *db;
	int count_index=list_index;
	for(db=buf_list.head; db && count_index; db=db->next,count_index--);
	for(; db; db=db->next)
	{
		++list_index;
		if(!strncmp(db->name,text,len))
		{  return(my_strdup(db->name));  }
	}
	
	/* If no names matched, then return NULL. */
	rl_attempted_completion_over=1;
	return(NULL);
}

static char *_atmelprg_setvar_generator(const char *text,int state)
{
	static int list_indexA, list_indexB, len;
	
	/* If this is a new word to complete, initialize now.  This
	   includes saving the length of TEXT for efficiency, and
	   initializing the index variable to 0. */
	if(!state)
	{
		list_indexA = 0;
		list_indexB = 0;
		len = strlen(text);
	}
	
	static const char *set_var_names_base[]=
	{ "help", "verbose", "idle_poll_msec", "busy_poll_usec", 
		"usb_resp_timeout", "usb_latency", "jtag_delay_fact", NULL };
	static const char *set_var_names_atmelprg[]=
	{ "spi_speed", "jtag_speed", "ext_reset", "try_clock_shift", NULL };
	static const char *set_var_names_uniprog[]=
	{ "spi_speed", "jtag_speed", "io_voltage", "sup_voltage", 
		"ext_reset", "try_clock_shift", NULL };
	
	/* Return the next name which partially matches from the
	   command list. */
	const char *name;
	while((name=set_var_names_base[list_indexA]))
	{
		list_indexA++;
		
		if(!strncmp(name,text,len))
		{  return(my_strdup(name));  }
	}
	
	const char **set_var_names_adapter=NULL;
	switch(USB_GetAdapterType())
	{
		case AT_AtmelPrg:
			set_var_names_adapter=set_var_names_atmelprg;
			break;
		case AT_UniProg:
			set_var_names_adapter=set_var_names_uniprog;
			break;
		//default: leave NULL. 
	}
	
	if(set_var_names_adapter)
	{
		while((name=set_var_names_adapter[list_indexB]))
		{
			list_indexB++;
			
			if(!strncmp(name,text,len))
			{  return(my_strdup(name));  }
		}
	}
	
	/* If no names matched, then return NULL. */
	rl_attempted_completion_over=1;
	return(NULL);
}

static char *_atmelprg_nomatch_generator(const char * /*text*/,int state)
{
	// There's nothing to complete...
	rl_attempted_completion_over=1;
	return(NULL);
}

static char **_atmelprg_readline_completion(const char *text,int start,
	int /*end*/)
{
	char **matches=NULL;
	
	// If this word is at the start of the line, then it is a command
	// to complete.  Otherwise it is the name of a file in the current
	// directory.
	// NOTE that text is the start of the word to complete (without leading 
	//      white space) while start is the index into the rl_line_buffer. 
	//      This makes sure that we complete on commands even if there are 
	//      white spaces at the beginning of the line. 
	const char *cur_buf=rl_line_buffer;
	// Simple count for argument number. 
	int arg_cnt=0;
	bool in_string=0;
	const char *last_cmd_start=cur_buf;  // For completion...
	for(int i=0; i<start; i++,cur_buf++)
	{
		if(*cur_buf==';')  // <-- Command separator. 
		{  arg_cnt=0;  in_string=0;  last_cmd_start=cur_buf+1;  }
		else
		{
			bool now_in_string=!isspace(*cur_buf);
			if(!in_string && now_in_string)
			{  ++arg_cnt;  }
			in_string=now_in_string;
		}
	}
	//fprintf(stderr,"<%d>\b\b\b",arg_cnt);
	
	// argument 0 is the command itself here. Complete on command. 
	if(arg_cnt==0)
	{  matches=rl_completion_matches(text,_atmelprg_command_generator);  }
	else
	{
		// Otherwise, perform completion depending on command. 
		const char *cmd=last_cmd_start;
		while(isspace(*cmd))  ++cmd;
		const char *end=cmd;
		while(*end && !isspace(*end))  ++end;
		char *cmd_copy=(char*)malloc(end-cmd+1);
		if(!cmd_copy)  return(matches);
		memcpy(cmd_copy,cmd,end-cmd);
		cmd_copy[end-cmd]='\0';
		
		//fprintf(stderr,"cmd=<%s>\n",cmd_copy);
		for(int i=0; commands[i].desc; i++)
		{
			if(!commands[i].name)  continue;
			if(strcmp(cmd_copy,commands[i].name))  continue;
			
			CompletionType ct=CT_None;
			     if(arg_cnt==1)  ct=commands[i].complete_1;
			else if(arg_cnt==2)  ct=commands[i].complete_2;
			switch(ct)
			{
				case CT_None:
					matches=rl_completion_matches(text,
						&_atmelprg_nomatch_generator);
					break;
				case CT_File:
					matches=NULL;
					break;
				case CT_Buffer:
					matches=rl_completion_matches(text,
						&_atmelprg_buffer_generator);
					break;
				case CT_SetVar:
					matches=rl_completion_matches(text,
						&_atmelprg_setvar_generator);
					break;
				default:  assert(0);
			}
			
			break;
		}
		
		free(cmd_copy);
	}
	
	return(matches);
}
#else /* No readline present. */
void NonReadlineReadChar()
{
	// NOTE: People should really use readline since we HAVE it!
	//       This replacement implementation sucks. It has a fixed 
	//       maximum command length. 
	static const int max_cmd_len=1024;
	static char cmd_buf[max_cmd_len];
	
	int rd=read(fileno(stdin),cmd_buf,max_cmd_len+1);
	if(rd<=0)
	{
		if(rd<0)
		{
			fprintf(stderr,"reading stdin: %s\n",strerror(errno));
			exit(1);
		}
		printf("quit\n");  fflush(stdout);
		++got_eof;
		return;
	}
	cmd_buf[rd]='\0';
	
	do {
		char *cmd=cmd_buf;
		if(*cmd=='\0')  break;
		
		while(isspace(*cmd))  ++cmd;
		
		char *buf_end=cmd+strlen(cmd);
		while(isspace(*--buf_end))  *buf_end='\0';
		
		ExecuteCommand(cmd);
	} while(0);
	
	printf(prompt_str);  fflush(stdout);
}
#endif


static void _ExitCleanup()
{
#if HAVE_LIBREADLINE
	rl_set_prompt("");
	rl_callback_handler_remove();
	rl_instream=NULL;
#endif
}


int main(int argc,char **arg)
{
	int vendor=FTDI_USB_VENDOR_CODE,device=FTDI_DEVICE_ATMELPRG;
	
	// Set prg_name to program name without path. 
	prg_name=strrchr(arg[0],'/');
	if(prg_name)  ++prg_name;
	else prg_name=arg[0];
	
	if(!strcasecmp(prg_name,"usb-atmelprg"))      device=FTDI_DEVICE_ATMELPRG;
	else if(!strcasecmp(prg_name,"usb-uniprog"))  device=FTDI_DEVICE_UNIPROG;
	else if(!strcasestr(prg_name,"atmelprg"))     device=FTDI_DEVICE_ATMELPRG;
	else if(!strcasestr(prg_name,"uniprog"))      device=FTDI_DEVICE_UNIPROG;
	
	int v_was_set=0,d_was_set=0;
	if(!ReadEnvVar_FTDI_USB(&vendor,&device))
	{  v_was_set=1;  d_was_set=1;  }
	
	// Process command line arguments. 
	int edit_ftdi_eeprom=0;
	for(int i=1; i<argc; i++)
	{
		if(!strncmp(arg[i],"-v=",3))
		{  vendor=strtol(arg[i]+3,NULL,16);  ++v_was_set;  }
		else if(!strncmp(arg[i],"-d=",3))
		{  device=strtol(arg[i]+3,NULL,16);  ++d_was_set;  }
		else if(!strcmp(arg[i],"-v"))
		{  ++verbose;  }
		else if(!strcmp(arg[i],"--version"))
		{
			_PrintVersion();
			exit(0);
		}
		else if(!strcmp(arg[i],"--help"))
		{  PrintUsage(0);  }
		else if(!strcmp(arg[i],"--edit-ftdi-eeprom"))
		{
			edit_ftdi_eeprom=1;
			if(!v_was_set)  vendor=FTDI_USB_VENDOR_CODE;
			if(!d_was_set)  device=FTDI_UNCONFIGURED_DEVICE;
		}
		else
		{  PrintUsage(1);  }
	}
	
	InstallSignalHandler(SIGINT,&_caught_sigint,0);
	
	// This will exit on failure. 
	USB_OpenConection(vendor,device);
	assert(usb_context);  // Should exit on failure. 
	
	// Set initial prompt. We will change this lateron. 
	DoSetPrompt();
	
	if(edit_ftdi_eeprom)
	{
		USB_EditFTDIEEPROM();
		exit(0);
	}
	
#if HAVE_LIBREADLINE
	using_history();
	// NOTE: Save/restore history using read_history(file), write_history(file). 
	
	// Tell the GNU Readline library how to complete.  We want to try to
	// complete on command names if this is the first word in the line, or
	// on filenames if not.
	/* Allow conditional parsing of the ~/.inputrc file. */
	rl_readline_name = "atmelprg";
	/* Tell the completer that we want a crack first. */
	rl_attempted_completion_function = &_atmelprg_readline_completion;
#endif
	
	// Make sure we clean up properly when the USB device is detached. 
	atexit(&_ExitCleanup);
	
	// Enter main loop: 
	printf("USB Programmer Host Software %s by Wolfgang Wieser.\n"
		"Try '?' for help, ^D to quit, ^C to interrupt.\n",VERSION);
#if HAVE_LIBREADLINE
	rl_instream=stdin;
	rl_callback_handler_install(/*prompt=*/prompt_str,
			/*lhandler=*/&_readline_callback);
	rl_redisplay();
#else
	printf(prompt_str);  fflush(stdout);
#endif
	
	for(;;)
	{
		if(caught_sigint)
		{  caught_sigint=0;  }
		
		struct pollfd pfd[1];
		pfd[0].fd=fileno(stdin);
		pfd[0].events=POLLIN;
		pfd[0].revents=0;
		int prv=poll(pfd,1,idle_poll_msec);
		if(prv<0)
		{
			if(errno!=EINTR)
			{
				fprintf(stderr,"poll: %s\n",strerror(errno));
				break;
			}
		}
		if(prv>0 && (pfd[0].revents))  // (POLLHUP|POLLIN)
		{
#if HAVE_LIBREADLINE
			rl_callback_read_char();
#else
			NonReadlineReadChar();
#endif
		}
		
		if(got_eof)
		{
#if HAVE_JTAG_SYSTEM
			if(opmode==JTAGSYS)
			{  LeaveJTAGMode();  got_eof=0;  }
			else break;
#else
			break;
#endif
		}
		if(caught_sigint)
		{  caught_sigint=0;  }
		
		int rv=USB_PollRead();
		if(rv>0)
		{  fprintf(stderr,"Ignoring %d unexpected bytes.\n",rv);  }
		else if(rv<0)
		{
			// USB error. Try to re-connect after 1 second. 
			fprintf(stderr,"Lost USB communication. "
				"Trying to re-connect in a second...\n");
			USB_CloseConnection();
			assert(!usb_context);
			
			usleep(1000*1000);
			
			USB_OpenConection(vendor,device);
			assert(usb_context);  // Should exit on failure. 
			
#if HAVE_LIBREADLINE
			rl_forced_update_display();
#endif
		}
	}
	
#if HAVE_LIBREADLINE
	rl_callback_handler_remove();
	rl_instream=NULL;
#endif
	
	USB_CloseConnection();
	
	return(0);
}
